"""
Swin Transformer v2, adapted from
https://github.com/pytorch/vision/blob/main/torchvision/models/swin_transformer.py
https://github.com/microsoft/Swin-Transformer/blob/main/models/swin_transformer_v2.py

Paper "Swin Transformer V2: Scaling Up Capacity and Resolution", https://arxiv.org/abs/2111.09883

Changes from original:
* Window size based on image size (image size // 32)
"""

# Reference license: BSD 3-Clause and MIT

from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import Permute
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType
from birder.net.swin_transformer_v1 import get_relative_position_bias
from birder.net.swin_transformer_v1 import patch_merging_pad
from birder.net.swin_transformer_v1 import shifted_window_attention


class PatchMerging(nn.Module):
    def __init__(self, dim: int) -> None:
        super().__init__()
        self.dim = dim
        self.reduction = nn.Linear(4 * dim, 2 * dim, bias=False)
        self.norm = nn.LayerNorm(2 * dim, eps=1e-5)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = patch_merging_pad(x)
        x = self.reduction(x)  # ... H/2 W/2 2*C
        x = self.norm(x)

        return x


class ShiftedWindowAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        window_size: tuple[int, int],
        shift_size: tuple[int, int],
        num_heads: int,
        qkv_bias: bool,
        proj_bias: bool,
    ) -> None:
        super().__init__()
        if len(window_size) != 2 or len(shift_size) != 2:
            raise ValueError("window_size and shift_size must be of length 2")

        self.window_size = window_size
        self.shift_size = shift_size
        self.num_heads = num_heads

        self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim, bias=proj_bias)

        self.define_relative_position_bias_table()
        self.define_relative_position_index()

        self.logit_scale = nn.Parameter(torch.log(10 * torch.ones((num_heads, 1, 1))), requires_grad=True)

        # MLP to generate continuous relative position bias
        self.cpb_mlp = nn.Sequential(
            nn.Linear(2, 512, bias=True), nn.ReLU(inplace=True), nn.Linear(512, num_heads, bias=False)
        )
        if qkv_bias is True:
            length = self.qkv.bias.numel() // 3
            self.qkv.bias[length : 2 * length].data.zero_()

    def define_relative_position_bias_table(self) -> None:
        # Get relative_coords_table
        relative_coords_h = torch.arange(-(self.window_size[0] - 1), self.window_size[0], dtype=torch.float32)
        relative_coords_w = torch.arange(-(self.window_size[1] - 1), self.window_size[1], dtype=torch.float32)
        relative_coords_table = torch.stack(torch.meshgrid([relative_coords_h, relative_coords_w], indexing="ij"))
        relative_coords_table = relative_coords_table.permute(1, 2, 0).contiguous().unsqueeze(0)  # 1, 2*Wh-1, 2*Ww-1, 2

        relative_coords_table[:, :, :, 0] /= self.window_size[0] - 1
        relative_coords_table[:, :, :, 1] /= self.window_size[1] - 1

        relative_coords_table *= 8  # Normalize to (-8, 8)
        relative_coords_table = (
            torch.sign(relative_coords_table) * torch.log2(torch.abs(relative_coords_table) + 1.0) / 3.0
        )
        self.relative_coords_table = nn.Buffer(relative_coords_table)

    def define_relative_position_index(self) -> None:
        # Get pair-wise relative position index for each token inside the window
        coords_h = torch.arange(self.window_size[0])
        coords_w = torch.arange(self.window_size[1])
        coords = torch.stack(torch.meshgrid(coords_h, coords_w, indexing="ij"))  # 2, Wh, Ww
        coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
        relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 2, Wh*Ww, Wh*Ww
        relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
        relative_coords[:, :, 0] += self.window_size[0] - 1  # Shift to start from 0
        relative_coords[:, :, 1] += self.window_size[1] - 1
        relative_coords[:, :, 0] *= 2 * self.window_size[1] - 1
        relative_position_index = relative_coords.sum(-1).flatten()  # Wh*Ww*Wh*Ww
        self.relative_position_index = nn.Buffer(relative_position_index)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        relative_position_bias = get_relative_position_bias(
            self.cpb_mlp(self.relative_coords_table).view(-1, self.num_heads),
            self.relative_position_index,
            self.window_size,
        )
        relative_position_bias = 16 * torch.sigmoid(relative_position_bias)

        return shifted_window_attention(
            x,
            self.qkv.weight,
            self.proj.weight,
            relative_position_bias,
            self.window_size,
            self.num_heads,
            shift_size=self.shift_size,
            qkv_bias=self.qkv.bias,
            proj_bias=self.proj.bias,
            logit_scale=self.logit_scale,
        )


class SwinTransformerBlock(nn.Module):
    def __init__(
        self,
        dim: int,
        input_resolution: tuple[int, int],
        num_heads: int,
        window_size: tuple[int, int],
        shift_size: tuple[int, int],
        mlp_ratio: float,
        stochastic_depth_prob: float,
    ) -> None:
        super().__init__()

        self.input_resolution = input_resolution
        window_size_h = window_size[0]
        window_size_w = window_size[1]
        shift_size_h = shift_size[0]
        shift_size_w = shift_size[1]
        if self.input_resolution[0] <= window_size_h:
            shift_size_h = 0
            window_size_h = self.input_resolution[0]
        if self.input_resolution[1] <= window_size_w:
            shift_size_w = 0
            window_size_w = self.input_resolution[1]

        window_size = (window_size_h, window_size_w)
        shift_size = (shift_size_h, shift_size_w)

        self.norm1 = nn.LayerNorm(dim, eps=1e-5)
        self.attn = ShiftedWindowAttention(
            dim,
            window_size,
            shift_size,
            num_heads,
            qkv_bias=True,
            proj_bias=True,
        )
        self.stochastic_depth = StochasticDepth(stochastic_depth_prob, "row")
        self.norm2 = nn.LayerNorm(dim, eps=1e-5)
        self.mlp = MLP(dim, [int(dim * mlp_ratio), dim], activation_layer=nn.GELU, inplace=None)

        for m in self.mlp.modules():
            if isinstance(m, nn.Linear):
                nn.init.xavier_uniform_(m.weight)
                if m.bias is not None:
                    nn.init.normal_(m.bias, std=1e-6)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.stochastic_depth(self.norm1(self.attn(x)))
        x = x + self.stochastic_depth(self.norm2(self.mlp(x)))

        return x


# pylint: disable=invalid-name
class Swin_Transformer_v2(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    default_size = (256, 256)
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    # pylint: disable=too-many-locals
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        patch_size = 4
        embed_dim: int = self.config["embed_dim"]
        depths: list[int] = self.config["depths"]
        num_heads: list[int] = self.config["num_heads"]
        drop_path_rate: float = self.config["drop_path_rate"]
        self.window_scale_factor: int = self.config["window_scale_factor"]
        mlp_ratio = 4.0
        window_size_h = int(self.size[0] / (2**5)) * self.window_scale_factor
        window_size_w = int(self.size[1] / (2**5)) * self.window_scale_factor
        window_size = (window_size_h, window_size_w)

        self.stem = nn.Sequential(
            nn.Conv2d(
                self.input_channels,
                embed_dim,
                kernel_size=(patch_size, patch_size),
                stride=patch_size,
                padding=(0, 0),
                bias=True,
            ),
            Permute([0, 2, 3, 1]),
            nn.LayerNorm(embed_dim, eps=1e-5),
        )

        resolution = (self.size[0] // patch_size, self.size[1] // patch_size)
        total_stage_blocks = sum(depths)
        stage_block_id = 0
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        layers = []
        for i_stage, depth in enumerate(depths):
            dim = embed_dim * 2**i_stage
            for i_layer in range(depth):
                # Adjust stochastic depth probability based on the depth of the stage block
                sd_prob = drop_path_rate * float(stage_block_id) / (total_stage_blocks - 1)
                if i_layer % 2 == 0:
                    shift_size = (0, 0)
                else:
                    shift_size = (window_size[0] // 2, window_size[1] // 2)

                layers.append(
                    SwinTransformerBlock(
                        dim,
                        resolution,
                        num_heads[i_stage],
                        window_size=window_size,
                        shift_size=shift_size,
                        mlp_ratio=mlp_ratio,
                        stochastic_depth_prob=sd_prob,
                    )
                )
                stage_block_id += 1

            stages[f"stage{i_stage+1}"] = nn.Sequential(*layers)
            return_channels.append(dim)
            layers = []

            # Add patch merging layer
            if i_stage < (len(depths) - 1):
                layers.append(PatchMerging(dim))
                resolution = (resolution[0] // 2, resolution[1] // 2)

        num_features = embed_dim * 2 ** (len(depths) - 1)
        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.LayerNorm(num_features, eps=1e-5),
            Permute([0, 3, 1, 2]),  # B H W C -> B C H W
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = num_features
        self.classifier = self.create_classifier()

        self.stem_stride = patch_size
        self.stem_width = embed_dim
        self.encoding_size = num_features

        # Weight initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, mean=0.0, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x.permute(0, 3, 1, 2).contiguous()

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(
            x, mask, channels_last=True, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token
        )
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x.permute(0, 3, 1, 2).contiguous()
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        old_size = self.size
        super().adjust_size(new_size)

        for m in self.body.modules():
            if isinstance(m, SwinTransformerBlock):
                new_window_size_h = int(new_size[0] / (2**5)) * self.window_scale_factor
                new_window_size_w = int(new_size[1] / (2**5)) * self.window_scale_factor
                new_window_size = (new_window_size_h, new_window_size_w)

                shift_size_h = m.attn.shift_size[0]
                shift_size_w = m.attn.shift_size[1]
                window_size_h = new_window_size[0]
                window_size_w = new_window_size[1]

                # Adjust resolution
                scale_h = old_size[0] // m.input_resolution[0]
                scale_w = old_size[1] // m.input_resolution[1]
                m.input_resolution = (new_size[0] // scale_h, new_size[1] // scale_w)

                if m.input_resolution[0] <= window_size_h:
                    shift_size_h = 0
                    window_size_h = m.input_resolution[0]

                if m.input_resolution[1] <= window_size_w:
                    shift_size_w = 0
                    window_size_w = m.input_resolution[1]

                m.attn.window_size = (window_size_h, window_size_w)

                if m.attn.shift_size[0] != 0:
                    shift_size_h = m.attn.window_size[0] // 2

                if m.attn.shift_size[1] != 0:
                    shift_size_w = m.attn.window_size[1] // 2

                m.attn.shift_size = (shift_size_h, shift_size_w)

                m.attn.define_relative_position_bias_table()
                m.attn.define_relative_position_index()


# Window factor = 1
registry.register_model_config(
    "swin_transformer_v2_t",
    Swin_Transformer_v2,
    config={
        "embed_dim": 96,
        "depths": [2, 2, 6, 2],
        "num_heads": [3, 6, 12, 24],
        "drop_path_rate": 0.2,
        "window_scale_factor": 1,
    },
)
registry.register_model_config(
    "swin_transformer_v2_s",
    Swin_Transformer_v2,
    config={
        "embed_dim": 96,
        "depths": [2, 2, 18, 2],
        "num_heads": [3, 6, 12, 24],
        "drop_path_rate": 0.3,
        "window_scale_factor": 1,
    },
)
registry.register_model_config(
    "swin_transformer_v2_b",
    Swin_Transformer_v2,
    config={
        "embed_dim": 128,
        "depths": [2, 2, 18, 2],
        "num_heads": [4, 8, 16, 32],
        "drop_path_rate": 0.5,
        "window_scale_factor": 1,
    },
)
registry.register_model_config(
    "swin_transformer_v2_l",
    Swin_Transformer_v2,
    config={
        "embed_dim": 192,
        "depths": [2, 2, 18, 2],
        "num_heads": [6, 12, 24, 48],
        "drop_path_rate": 0.5,
        "window_scale_factor": 1,
    },
)

# Window factor = 2
registry.register_model_config(
    "swin_transformer_v2_w2_t",
    Swin_Transformer_v2,
    config={
        "embed_dim": 96,
        "depths": [2, 2, 6, 2],
        "num_heads": [3, 6, 12, 24],
        "drop_path_rate": 0.2,
        "window_scale_factor": 2,
    },
)
registry.register_model_config(
    "swin_transformer_v2_w2_s",
    Swin_Transformer_v2,
    config={
        "embed_dim": 96,
        "depths": [2, 2, 18, 2],
        "num_heads": [3, 6, 12, 24],
        "drop_path_rate": 0.3,
        "window_scale_factor": 2,
    },
)
registry.register_model_config(
    "swin_transformer_v2_w2_b",
    Swin_Transformer_v2,
    config={
        "embed_dim": 128,
        "depths": [2, 2, 18, 2],
        "num_heads": [4, 8, 16, 32],
        "drop_path_rate": 0.5,
        "window_scale_factor": 2,
    },
)
registry.register_model_config(
    "swin_transformer_v2_w2_l",
    Swin_Transformer_v2,
    config={
        "embed_dim": 192,
        "depths": [2, 2, 18, 2],
        "num_heads": [6, 12, 24, 48],
        "drop_path_rate": 0.5,
        "window_scale_factor": 2,
    },
)

registry.register_weights(
    "swin_transformer_v2_s_intermediate-arabian-peninsula",
    {
        "url": (
            "https://huggingface.co/birder-project/swin_transformer_v2_s_intermediate-arabian-peninsula/resolve/main"
        ),
        "description": (
            "Swin Transformer v2 small model with intermediate training, "
            "then fine-tuned on the arabian-peninsula dataset"
        ),
        "resolution": (384, 384),
        "formats": {
            "pt": {
                "file_size": 193.0,
                "sha256": "766706779aef6b55860cabd5712bcb758f463896100645648f917495214dd478",
            },
        },
        "net": {"network": "swin_transformer_v2_s", "tag": "intermediate-arabian-peninsula"},
    },
)



import math
from typing import Any

import torch


def collate_fn(batch: list[tuple[Any, ...]]) -> tuple[Any, ...]:
    return tuple(zip(*batch))


def batch_images(images: list[torch.Tensor], size_divisible: int) -> tuple[torch.Tensor, torch.Tensor, list[list[int]]]:
    """
    Batch list of image tensors of different sizes into a single batch.
    Pad with zeros all images to the shape of the largest image in the list.
    """

    size_list = [list(img.shape) for img in images]
    max_size = size_list[0]
    for sublist in size_list[1:]:
        for index, item in enumerate(sublist):
            max_size[index] = max(max_size[index], item)

    image_sizes = [img.shape[-2:] for img in images]
    stride = float(size_divisible)
    max_size = list(max_size)
    max_size[1] = int(math.ceil(float(max_size[1]) / stride) * stride)
    max_size[2] = int(math.ceil(float(max_size[2]) / stride) * stride)

    batch_shape = [len(images)] + max_size
    (B, _, H, W) = batch_shape
    batched_imgs = images[0].new_full(batch_shape, 0)
    masks = images[0].new_full((B, H, W), 1).to(torch.bool)
    for img, pad_img, mask in zip(images, batched_imgs, masks):
        pad_img[: img.shape[0], : img.shape[1], : img.shape[2]].copy_(img)
        mask[: img.shape[1], : img.shape[2]] = False

    return (batched_imgs, masks, image_sizes)


class DetectionCollator:
    def __init__(self, input_offset: int, size_divisible: int = 32) -> None:
        self.offset = input_offset
        self.size_divisible = size_divisible

    def __call__(self, batch: list[tuple[Any, ...]]) -> tuple[Any, ...]:
        # In case that the "transforms" returning a tuple, flatten it
        # if isinstance(batch[0][self.offset], tuple):
        #     batch = [(*sample[: self.offset], *sample[self.offset], *sample[self.offset + 1 :]) for sample in batch]

        data = collate_fn(batch)
        images: list[torch.Tensor] = data[self.offset]
        assert images[0].ndim == 3

        (batched_imgs, masks, size_list) = batch_images(images, self.size_divisible)

        return data[: self.offset] + (batched_imgs,) + data[self.offset + 1 :] + (masks, size_list)


# inputs, targets
training_collate_fn = DetectionCollator(0)

# file_paths, inputs, targets
inference_collate_fn = DetectionCollator(1)



import argparse
import logging
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import torch
from PIL import Image

from birder.adversarial.fgsm import FGSM
from birder.adversarial.pgd import PGD
from birder.common import cli
from birder.common import fs_ops
from birder.common import lib
from birder.data.transforms.classification import inference_preset
from birder.data.transforms.classification import reverse_preset

logger = logging.getLogger(__name__)


def show_pgd(args: argparse.Namespace) -> None:
    if args.gpu is True:
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    if args.gpu_id is not None:
        torch.cuda.set_device(args.gpu_id)

    logger.info(f"Using device {device}")

    (net, (class_to_idx, signature, rgb_stats, *_)) = fs_ops.load_model(
        device,
        args.network,
        tag=args.tag,
        epoch=args.epoch,
        inference=True,
        reparameterized=args.reparameterized,
    )
    label_names = list(class_to_idx.keys())
    size = lib.get_size_from_signature(signature)
    transform = inference_preset(size, rgb_stats, 1.0)
    reverse_transform = reverse_preset(rgb_stats)

    img: Image.Image = Image.open(args.image_path)
    input_tensor = transform(img).unsqueeze(dim=0).to(device)

    pgd = PGD(net, eps=args.eps, max_delta=0.012, steps=10, random_start=True)
    if args.target is not None:
        target = torch.tensor(class_to_idx[args.target]).unsqueeze(dim=0).to(device)
    else:
        target = None

    img = img.resize(size)
    pgd_response = pgd(input_tensor, target=target)
    perturbation = reverse_transform(pgd_response.adv_img).cpu().detach().numpy().squeeze()
    pgd_img = np.moveaxis(perturbation, 0, 2)

    # Get predictions and probabilities
    prob = pgd_response.out.cpu().detach().numpy().squeeze()
    adv_prob = pgd_response.adv_out.cpu().detach().numpy().squeeze()
    idx = np.argmax(prob)
    adv_idx = np.argmax(adv_prob)

    _, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8))
    ax1.imshow(img)
    ax1.set_title(f"{label_names[idx]} {100 * prob[idx]:.2f}%")
    ax2.imshow(pgd_img)
    ax2.set_title(f"{label_names[adv_idx]} {100 * adv_prob[adv_idx]:.2f}%")
    plt.show()


def show_fgsm(args: argparse.Namespace) -> None:
    if args.gpu is True:
        device = torch.device("cuda")
    else:
        device = torch.device("cpu")

    if args.gpu_id is not None:
        torch.cuda.set_device(args.gpu_id)

    logger.info(f"Using device {device}")

    (net, (class_to_idx, signature, rgb_stats, *_)) = fs_ops.load_model(
        device,
        args.network,
        tag=args.tag,
        epoch=args.epoch,
        inference=False,
        reparameterized=args.reparameterized,
    )
    label_names = list(class_to_idx.keys())
    size = lib.get_size_from_signature(signature)
    transform = inference_preset(size, rgb_stats, 1.0)

    img: Image.Image = Image.open(args.image_path)
    input_tensor = transform(img).unsqueeze(dim=0).to(device)

    fgsm = FGSM(net, eps=args.eps)
    if args.target is not None:
        target = torch.tensor(class_to_idx[args.target]).unsqueeze(dim=0).to(device)
    else:
        target = None

    img = img.resize(size)
    fgsm_response = fgsm(input_tensor, target=target)
    perturbation = fgsm_response.perturbation.cpu().detach().numpy().squeeze()
    fgsm_img = (np.array(img).astype(np.float32) / 255.0) + np.moveaxis(perturbation, 0, 2)
    fgsm_img = np.clip(fgsm_img, 0, 1)

    # Get predictions and probabilities
    prob = fgsm_response.out.cpu().detach().numpy().squeeze()
    adv_prob = fgsm_response.adv_out.cpu().detach().numpy().squeeze()
    idx = np.argmax(prob)
    adv_idx = np.argmax(adv_prob)

    _, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8))
    ax1.imshow(img)
    ax1.set_title(f"{label_names[idx]} {100 * prob[idx]:.2f}%")
    ax2.imshow(fgsm_img)
    ax2.set_title(f"{label_names[adv_idx]} {100 * adv_prob[adv_idx]:.2f}%")
    plt.show()


def set_parser(subparsers: Any) -> None:
    subparser = subparsers.add_parser(
        "adversarial",
        allow_abbrev=False,
        help="deep learning adversarial attacks",
        description="deep learning adversarial attacks",
        epilog=(
            "Usage examples:\n"
            "python -m birder.tools adversarial --method fgsm --network efficientnet_v2_s "
            "--epoch 0 --target Bluethroat 'data/training/Mallard/000117.jpeg'\n"
            "python -m birder.tools adversarial --method fgsm --network efficientnet_v2_m "
            "--epoch 0 --eps 0.02 --target Mallard 'data/validation/White-tailed eagle/000006.jpeg'\n"
            "python tool.py adversarial --method pgd --network caformer_s18 -e 0 "
            "data/validation/Arabian babbler/000001.jpeg\n"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    subparser.add_argument("--method", type=str, choices=["fgsm", "pgd"], help="introspection method")
    subparser.add_argument(
        "-n", "--network", type=str, required=True, help="the neural network to use (i.e. resnet_v2)"
    )
    subparser.add_argument("-e", "--epoch", type=int, metavar="N", help="model checkpoint to load")
    subparser.add_argument("-t", "--tag", type=str, help="model tag (from the training phase)")
    subparser.add_argument(
        "-r", "--reparameterized", default=False, action="store_true", help="load reparameterized model"
    )
    subparser.add_argument("--gpu", default=False, action="store_true", help="use gpu")
    subparser.add_argument("--gpu-id", type=int, metavar="ID", help="gpu id to use")
    subparser.add_argument("--eps", type=float, default=0.007, help="fgsm epsilon")
    subparser.add_argument("--target", type=str, help="target class, leave empty to use predicted class")
    subparser.add_argument("image_path", type=str, help="input image path")
    subparser.set_defaults(func=main)


def main(args: argparse.Namespace) -> None:
    if args.method == "fgsm":
        show_fgsm(args)
    elif args.method == "pgd":
        show_pgd(args)



"""
DETR (DEtection TRansformer), adapted from
https://github.com/facebookresearch/detr/blob/main/models/detr.py

Paper "End-to-End Object Detection with Transformers", https://arxiv.org/abs/2005.12872

Changes from original:
* Move background index to first from last (to be inline with the rest of Birder detectors)
* Zero cost matrix elements on overflow (HungarianMatcher)
"""

# Reference license: Apache-2.0

import copy
import math
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from scipy.optimize import linear_sum_assignment
from torch import nn
from torchvision.ops import MLP
from torchvision.ops import boxes as box_ops

from birder.common import training_utils
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.detection.base import DetectionBaseNet
from birder.ops.soft_nms import SoftNMS


def _get_clones(module: nn.Module, N: int) -> nn.ModuleList:
    return nn.ModuleList([copy.deepcopy(module) for _ in range(N)])


class HungarianMatcher(nn.Module):
    """
    This class computes an assignment between the targets and the predictions of the network
    """

    def __init__(self, cost_class: float, cost_bbox: float, cost_giou: float):
        super().__init__()
        assert cost_class != 0 or cost_bbox != 0 or cost_giou != 0
        self.cost_class = cost_class
        self.cost_bbox = cost_bbox
        self.cost_giou = cost_giou

    @torch.jit.unused  # type: ignore[misc]
    def forward(
        self, class_logits: torch.Tensor, box_regression: torch.Tensor, targets: list[dict[str, torch.Tensor]]
    ) -> list[torch.Tensor]:
        with torch.no_grad():
            (B, num_queries) = class_logits.shape[:2]

            # We flatten to compute the cost matrices in a batch
            out_prob = class_logits.flatten(0, 1).softmax(-1)  # [batch_size * num_queries, num_classes]
            out_bbox = box_regression.flatten(0, 1)  # [batch_size * num_queries, 4]

            # Also concat the target labels and boxes
            tgt_ids = torch.concat([v["labels"] for v in targets], dim=0)
            tgt_bbox = torch.concat([v["boxes"] for v in targets], dim=0)

            # Compute the classification cost. Contrary to the loss, we don't use the NLL,
            # but approximate it in 1 - prob[target class].
            # The 1 is a constant that doesn't change the matching, it can be omitted.
            cost_class = -out_prob[:, tgt_ids]

            # Compute the L1 cost between boxes
            cost_bbox = torch.cdist(out_bbox, tgt_bbox, p=1.0)

            # Compute the giou cost between boxes
            cost_giou = -box_ops.generalized_box_iou(
                box_ops.box_convert(out_bbox, in_fmt="cxcywh", out_fmt="xyxy"),
                box_ops.box_convert(tgt_bbox, in_fmt="cxcywh", out_fmt="xyxy"),
            )

            # Final cost matrix
            C = self.cost_bbox * cost_bbox + self.cost_class * cost_class + self.cost_giou * cost_giou
            C = C.view(B, num_queries, -1).cpu()
            C[C.isnan() | C.isinf()] = 0.0

            sizes = [len(v["boxes"]) for v in targets]
            indices = [linear_sum_assignment(c[i]) for i, c in enumerate(C.split(sizes, -1))]

            return [(torch.as_tensor(i, dtype=torch.int64), torch.as_tensor(j, dtype=torch.int64)) for i, j in indices]


class TransformerEncoderLayer(nn.Module):
    def __init__(self, d_model: int, num_heads: int, dim_feedforward: int, dropout: float) -> None:
        super().__init__()
        self.self_attn = nn.MultiheadAttention(d_model, num_heads, dropout=dropout)

        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)

        self.activation = nn.ReLU()

    def forward(
        self, src: torch.Tensor, pos: torch.Tensor, src_key_padding_mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        q = src + pos
        k = src + pos

        (src2, _) = self.self_attn(q, k, value=src, key_padding_mask=src_key_padding_mask)
        src = src + self.dropout1(src2)
        src = self.norm1(src)
        src2 = self.linear2(self.dropout(self.activation(self.linear1(src))))
        src = src + self.dropout2(src2)
        src = self.norm2(src)

        return src


class TransformerDecoderLayer(nn.Module):
    def __init__(self, d_model: int, num_heads: int, dim_feedforward: int, dropout: float) -> None:
        super().__init__()
        self.self_attn = nn.MultiheadAttention(d_model, num_heads, dropout=dropout)
        self.multihead_attn = nn.MultiheadAttention(d_model, num_heads, dropout=dropout)

        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.norm3 = nn.LayerNorm(d_model)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)
        self.dropout3 = nn.Dropout(dropout)

        self.activation = nn.ReLU()

    def forward(
        self,
        tgt: torch.Tensor,
        memory: torch.Tensor,
        pos: torch.Tensor,
        query_pos: torch.Tensor,
        memory_key_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        q = tgt + query_pos
        k = tgt + query_pos

        (tgt2, _) = self.self_attn(q, k, value=tgt)
        tgt = tgt + self.dropout1(tgt2)
        tgt = self.norm1(tgt)
        (tgt2, _) = self.multihead_attn(
            query=tgt + query_pos, key=memory + pos, value=memory, key_padding_mask=memory_key_padding_mask
        )
        tgt = tgt + self.dropout2(tgt2)
        tgt = self.norm2(tgt)
        tgt2 = self.linear2(self.dropout(self.activation(self.linear1(tgt))))
        tgt = tgt + self.dropout3(tgt2)
        tgt = self.norm3(tgt)

        return tgt


class TransformerEncoder(nn.Module):
    def __init__(self, encoder_layer: nn.Module, num_layers: int) -> None:
        super().__init__()
        self.layers = _get_clones(encoder_layer, num_layers)

    def forward(
        self, x: torch.Tensor, pos: torch.Tensor, x_key_padding_mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        out = x
        for layer in self.layers:
            out = layer(out, pos=pos, src_key_padding_mask=x_key_padding_mask)

        return out


class TransformerDecoder(nn.Module):
    def __init__(self, decoder_layer: nn.Module, num_layers: int, norm: nn.Module, return_intermediate: bool) -> None:
        super().__init__()
        self.layers = _get_clones(decoder_layer, num_layers)
        self.num_layers = num_layers
        self.norm = norm
        self.return_intermediate = return_intermediate

    def forward(
        self,
        tgt: torch.Tensor,
        memory: torch.Tensor,
        pos: torch.Tensor,
        query_pos: torch.Tensor,
        memory_key_padding_mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        output = tgt
        intermediate = []
        for layer in self.layers:
            output = layer(
                output, memory, pos=pos, query_pos=query_pos, memory_key_padding_mask=memory_key_padding_mask
            )
            if self.return_intermediate is True:
                intermediate.append(self.norm(output))

        if self.return_intermediate is True:
            return torch.stack(intermediate)

        return self.norm(output).unsqueeze(0)


class Transformer(nn.Module):
    def __init__(
        self,
        d_model: int,
        num_heads: int,
        num_encoder_layers: int,
        num_decoder_layers: int,
        dim_feedforward: int,
        dropout: float,
        return_intermediate_dec: bool,
    ) -> None:
        super().__init__()
        encoder_layer = TransformerEncoderLayer(d_model, num_heads, dim_feedforward, dropout)
        self.encoder = TransformerEncoder(encoder_layer, num_encoder_layers)

        decoder_layer = TransformerDecoderLayer(d_model, num_heads, dim_feedforward, dropout)
        decoder_norm = nn.LayerNorm(d_model)
        self.decoder = TransformerDecoder(
            decoder_layer, num_decoder_layers, decoder_norm, return_intermediate=return_intermediate_dec
        )

        # Weights initialization
        for p in self.parameters():
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def forward(
        self, src: torch.Tensor, query_embed: torch.Tensor, pos_embed: torch.Tensor, mask: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        # Flatten BxCxHxW to HWxBxC
        B = src.size(0)
        src = src.flatten(2).permute(2, 0, 1)
        pos_embed = pos_embed.flatten(2).permute(2, 0, 1)
        query_embed = query_embed.unsqueeze(1).repeat(1, B, 1)
        if mask is not None:
            mask = mask.flatten(1)

        tgt = torch.zeros_like(query_embed)
        memory = self.encoder(src, pos=pos_embed, x_key_padding_mask=mask)
        hs = self.decoder(tgt, memory, pos=pos_embed, query_pos=query_embed, memory_key_padding_mask=mask)

        return hs.transpose(1, 2)


class PositionEmbeddingSine(nn.Module):
    def __init__(self, num_pos_feats: int, temperature: int = 10000, normalize: bool = False) -> None:
        super().__init__()
        self.num_pos_feats = num_pos_feats
        self.temperature = temperature
        self.normalize = normalize
        self.scale = 2 * math.pi

    def forward(self, x: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        if mask is None:
            (B, _, H, W) = x.size()
            mask = torch.zeros(B, H, W, dtype=torch.bool, device=x.device)

        not_mask = ~mask
        y_embed = not_mask.cumsum(1, dtype=torch.float32)
        x_embed = not_mask.cumsum(2, dtype=torch.float32)
        if self.normalize is True:
            eps = 1e-6
            y_embed = y_embed / (y_embed[:, -1:, :] + eps) * self.scale
            x_embed = x_embed / (x_embed[:, :, -1:] + eps) * self.scale

        dim_t = torch.arange(self.num_pos_feats, dtype=torch.float32, device=x.device)
        dim_t = self.temperature ** (2 * (dim_t // 2) / self.num_pos_feats)

        pos_x = x_embed[:, :, :, None] / dim_t
        pos_y = y_embed[:, :, :, None] / dim_t
        pos_x = torch.stack((pos_x[:, :, :, 0::2].sin(), pos_x[:, :, :, 1::2].cos()), dim=4).flatten(3)
        pos_y = torch.stack((pos_y[:, :, :, 0::2].sin(), pos_y[:, :, :, 1::2].cos()), dim=4).flatten(3)
        pos = torch.concat((pos_y, pos_x), dim=3).permute(0, 3, 1, 2)

        return pos


class DETR(DetectionBaseNet):
    default_size = (640, 640)

    def __init__(
        self,
        num_classes: int,
        backbone: DetectorBackbone,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
        export_mode: bool = False,
    ) -> None:
        super().__init__(num_classes, backbone, config=config, size=size, export_mode=export_mode)
        assert self.config is not None, "must set config"

        hidden_dim = 256
        num_heads = 8
        dim_feedforward = 2048
        dropout: float = self.config.get("dropout", 0.1)
        num_encoder_layers: int = self.config["num_encoder_layers"]
        num_decoder_layers: int = self.config["num_decoder_layers"]
        num_queries: int = self.config.get("num_queries", 100)
        return_intermediate: bool = self.config.get("return_intermediate", True)
        soft_nms: bool = self.config.get("soft_nms", False)

        self.soft_nms = None
        if soft_nms is True:
            self.soft_nms = SoftNMS()

        self.hidden_dim = hidden_dim
        self.transformer = Transformer(
            d_model=hidden_dim,
            num_heads=num_heads,
            num_encoder_layers=num_encoder_layers,
            num_decoder_layers=num_decoder_layers,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            return_intermediate_dec=return_intermediate,
        )

        self.class_embed = nn.Linear(hidden_dim, self.num_classes)
        self.bbox_embed = MLP(hidden_dim, [hidden_dim, hidden_dim, 4], activation_layer=nn.ReLU)
        self.query_embed = nn.Embedding(num_queries, hidden_dim)
        self.input_proj = nn.Conv2d(
            self.backbone.return_channels[-1], hidden_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0)
        )
        self.pos_enc = PositionEmbeddingSine(hidden_dim // 2, normalize=True)

        self.matcher = HungarianMatcher(cost_class=1, cost_bbox=5, cost_giou=2)
        empty_weight = torch.ones(self.num_classes)
        empty_weight[0] = 0.1
        self.empty_weight = nn.Buffer(empty_weight)

        if self.export_mode is False:
            self.forward = torch.compiler.disable(recursive=False)(self.forward)  # type: ignore[method-assign]

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes + 1
        self.class_embed = nn.Linear(self.hidden_dim, self.num_classes)

        empty_weight = torch.ones(self.num_classes)
        empty_weight[0] = 0.1
        self.empty_weight = nn.Buffer(empty_weight)

    def freeze(self, freeze_classifier: bool = True) -> None:
        for param in self.parameters():
            param.requires_grad = False

        if freeze_classifier is False:
            for param in self.class_embed.parameters():
                param.requires_grad = True

    def _get_src_permutation_idx(self, indices: list[torch.Tensor]) -> tuple[torch.Tensor, torch.Tensor]:
        batch_idx = torch.concat([torch.full_like(src, i) for i, (src, _) in enumerate(indices)])
        src_idx = torch.concat([src for (src, _) in indices])
        return (batch_idx, src_idx)

    def _class_loss(
        self,
        cls_logits: torch.Tensor,
        targets: list[dict[str, torch.Tensor]],
        indices: list[torch.Tensor],
    ) -> torch.Tensor:
        idx = self._get_src_permutation_idx(indices)
        target_classes_o = torch.concat([t["labels"][J] for t, (_, J) in zip(targets, indices)], dim=0)
        target_classes = torch.full(cls_logits.shape[:2], 0, dtype=torch.int64, device=cls_logits.device)
        target_classes[idx] = target_classes_o
        loss_ce = F.cross_entropy(cls_logits.transpose(1, 2), target_classes, self.empty_weight)

        return loss_ce

    def _box_loss(
        self,
        box_output: torch.Tensor,
        targets: list[dict[str, torch.Tensor]],
        indices: list[torch.Tensor],
        num_boxes: int,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        idx = self._get_src_permutation_idx(indices)
        src_boxes = box_output[idx]
        target_boxes = torch.concat([t["boxes"][i] for t, (_, i) in zip(targets, indices)], dim=0)

        loss_bbox = F.l1_loss(src_boxes, target_boxes, reduction="none")
        loss_bbox = loss_bbox.sum() / num_boxes

        loss_giou = 1 - torch.diag(
            box_ops.generalized_box_iou(
                box_ops.box_convert(src_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
                box_ops.box_convert(target_boxes, in_fmt="cxcywh", out_fmt="xyxy"),
            )
        )
        loss_giou = loss_giou.sum() / num_boxes

        return (loss_bbox, loss_giou)

    @torch.jit.unused  # type: ignore[misc]
    @torch.compiler.disable()  # type: ignore[misc]
    def compute_loss(
        self,
        targets: list[dict[str, torch.Tensor]],
        cls_logits: torch.Tensor,
        box_output: torch.Tensor,
    ) -> dict[str, torch.Tensor]:
        # Compute the average number of target boxes across all nodes, for normalization purposes
        num_boxes = sum(len(t["labels"]) for t in targets)
        num_boxes = torch.as_tensor([num_boxes], dtype=torch.float, device=cls_logits.device)
        if training_utils.is_dist_available_and_initialized() is True:
            torch.distributed.all_reduce(num_boxes)

        num_boxes = torch.clamp(num_boxes / training_utils.get_world_size(), min=1).item()

        loss_ce_list = []
        loss_bbox_list = []
        loss_giou_list = []
        for idx in range(cls_logits.size(0)):
            indices = self.matcher(cls_logits[idx], box_output[idx], targets)
            loss_ce_i = self._class_loss(cls_logits[idx], targets, indices)
            (loss_bbox_i, loss_giou_i) = self._box_loss(box_output[idx], targets, indices, num_boxes)
            loss_ce_list.append(loss_ce_i)
            loss_bbox_list.append(loss_bbox_i)
            loss_giou_list.append(loss_giou_i)

        loss_ce = torch.stack(loss_ce_list).sum() * 1
        loss_bbox = torch.stack(loss_bbox_list).sum() * 5
        loss_giou = torch.stack(loss_giou_list).sum() * 2
        losses = {
            "labels": loss_ce,
            "boxes": loss_bbox,
            "giou": loss_giou,
        }

        return losses

    def postprocess_detections(
        self, class_logits: torch.Tensor, box_regression: torch.Tensor, image_shapes: list[tuple[int, int]]
    ) -> list[dict[str, torch.Tensor]]:
        prob = F.softmax(class_logits, -1)
        (scores, labels) = prob[..., 1:].max(-1)
        labels = labels + 1

        # TorchScript doesn't support creating tensor from tuples, convert everything to lists
        target_sizes = torch.tensor([list(s) for s in image_shapes], device=class_logits.device)

        # Convert to [x0, y0, x1, y1] format
        boxes = box_ops.box_convert(box_regression, in_fmt="cxcywh", out_fmt="xyxy")

        # Convert from relative [0, 1] to absolute [0, height] coordinates
        (img_h, img_w) = target_sizes.unbind(1)
        scale_fct = torch.stack([img_w, img_h, img_w, img_h], dim=1)
        boxes = boxes * scale_fct[:, None, :]

        detections: list[dict[str, torch.Tensor]] = []
        for s, l, b in zip(scores, labels, boxes):
            # Non-maximum suppression
            if self.soft_nms is not None:
                # Actually much faster on CPU
                device = b.device
                (soft_scores, keep) = self.soft_nms(b.cpu(), s.cpu(), l.cpu(), score_threshold=0.001)
                keep = keep.to(device)
                s[keep] = soft_scores.to(device)

                b = b[keep]
                s = s[keep]
                l = l[keep]  # noqa: E741

            detections.append(
                {
                    "boxes": b,
                    "scores": s,
                    "labels": l,
                }
            )

        return detections

    def forward(
        self,
        x: torch.Tensor,
        targets: Optional[list[dict[str, torch.Tensor]]] = None,
        masks: Optional[torch.Tensor] = None,
        image_sizes: Optional[list[list[int]]] = None,
    ) -> tuple[list[dict[str, torch.Tensor]], dict[str, torch.Tensor]]:
        self._input_check(targets)
        images = self._to_img_list(x, image_sizes)

        features: dict[str, torch.Tensor] = self.backbone.detection_features(x)
        x = features[self.backbone.return_stages[-1]]
        if masks is not None:
            masks = F.interpolate(masks[None].float(), size=x.shape[-2:], mode="nearest").to(torch.bool)[0]

        pos = self.pos_enc(x, masks)
        hs = self.transformer(self.input_proj(x), self.query_embed.weight, pos_embed=pos, mask=masks)
        outputs_class = self.class_embed(hs)
        outputs_coord = self.bbox_embed(hs).sigmoid()

        losses = {}
        detections: list[dict[str, torch.Tensor]] = []
        if self.training is True:
            assert targets is not None, "targets should not be none when in training mode"

            # Convert target boxes
            for idx, target in enumerate(targets):
                boxes = target["boxes"]
                boxes = box_ops.box_convert(boxes, in_fmt="xyxy", out_fmt="cxcywh")
                boxes = boxes / torch.tensor(images.image_sizes[idx][::-1] * 2, dtype=torch.float32, device=x.device)
                targets[idx]["boxes"] = boxes

            losses = self.compute_loss(targets, outputs_class, outputs_coord)

        else:
            detections = self.postprocess_detections(outputs_class[-1], outputs_coord[-1], images.image_sizes)

        return (detections, losses)


registry.register_model_config("detr", DETR, config={"num_encoder_layers": 6, "num_decoder_layers": 6})



"""
Next-ViT, adapted from
https://github.com/bytedance/Next-ViT/blob/main/classification/nextvit.py

"Next-ViT: Next Generation Vision Transformer for Efficient Deployment in Realistic Industrial Scenarios",
https://arxiv.org/abs/2207.05501
"""

# Reference license: Apache-2.0

from collections import OrderedDict
from typing import Any
from typing import Literal
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import Conv2dNormActivation
from torchvision.ops import StochasticDepth

from birder.common.masking import mask_tensor
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.base import TokenRetentionResultType
from birder.net.base import make_divisible


class PatchEmbed(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, stride: tuple[int, int]) -> None:
        super().__init__()
        if stride[0] > 1 or stride[1] > 1:
            self.avg_pool = nn.AvgPool2d(kernel_size=(2, 2), stride=(2, 2), ceil_mode=True, count_include_pad=False)
            self.conv = nn.Conv2d(
                in_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False
            )
            self.norm = nn.BatchNorm2d(out_channels)

        elif in_channels != out_channels:
            self.avg_pool = nn.Identity()
            self.conv = nn.Conv2d(
                in_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False
            )
            self.norm = nn.BatchNorm2d(out_channels)

        else:
            self.avg_pool = nn.Identity()
            self.conv = nn.Identity()
            self.norm = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.norm(self.conv(self.avg_pool(x)))


class MHCA(nn.Module):
    """
    Multi-Head Convolutional Attention
    """

    def __init__(self, out_channels: int, head_dim: int) -> None:
        super().__init__()
        self.group_conv3x3 = nn.Conv2d(
            out_channels,
            out_channels,
            kernel_size=(3, 3),
            stride=(1, 1),
            padding=(1, 1),
            groups=out_channels // head_dim,
            bias=False,
        )
        self.norm = nn.BatchNorm2d(out_channels)
        self.act = nn.ReLU(inplace=True)
        self.projection = nn.Conv2d(
            out_channels, out_channels, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=False
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.group_conv3x3(x)
        x = self.norm(x)
        x = self.act(x)
        x = self.projection(x)

        return x


class MLP(nn.Module):
    def __init__(self, in_features: int, out_features: int, mlp_ratio: float, bias: bool) -> None:
        super().__init__()
        hidden_dim = make_divisible(in_features * mlp_ratio, 32)
        self.conv1 = nn.Conv2d(in_features, hidden_dim, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=bias)
        self.act = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(hidden_dim, out_features, kernel_size=(1, 1), stride=(1, 1), padding=(0, 0), bias=bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.conv1(x)
        x = self.act(x)
        x = self.conv2(x)

        return x


class NCB(nn.Module):
    """
    Next Convolution Block
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        stride: tuple[int, int],
        drop_path: float,
        head_dim: int,
        mlp_ratio: float,
    ) -> None:
        super().__init__()
        assert out_channels % head_dim == 0

        self.patch_embed = PatchEmbed(in_channels, out_channels, stride)
        self.mhca = MHCA(out_channels, head_dim)
        self.drop_path = StochasticDepth(drop_path, mode="row")

        self.norm = nn.BatchNorm2d(out_channels)
        self.mlp = MLP(out_channels, out_channels, mlp_ratio=mlp_ratio, bias=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.patch_embed(x)
        x = x + self.drop_path(self.mhca(x))
        x = x + self.drop_path(self.mlp(self.norm(x)))

        return x


# pylint: disable=invalid-name
class E_MHSA(nn.Module):
    """
    Efficient Multi-Head Self Attention
    """

    def __init__(
        self,
        dim: int,
        out_dim: int,
        head_dim: int,
        sr_ratio: int,
    ) -> None:
        super().__init__()
        self.num_heads = dim // head_dim
        self.scale = head_dim**-0.5

        self.q = nn.Linear(dim, dim)
        self.k = nn.Linear(dim, dim)
        self.v = nn.Linear(dim, dim)
        self.proj = nn.Linear(dim, out_dim)

        n_ratio = sr_ratio**2
        if sr_ratio > 1:
            self.sr = nn.AvgPool1d(kernel_size=n_ratio, stride=n_ratio)
            self.norm = nn.BatchNorm1d(dim)

        else:
            self.sr = nn.Identity()
            self.norm = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, C) = x.size()
        q = self.q(x)
        q = q.reshape(B, N, self.num_heads, int(C // self.num_heads)).permute(0, 2, 1, 3)

        x = x.transpose(1, 2)
        x = self.sr(x)
        x = self.norm(x)
        x = x.transpose(1, 2)

        k = self.k(x)
        k = k.reshape(B, -1, self.num_heads, int(C // self.num_heads)).transpose(1, 2)
        v = self.v(x)
        v = v.reshape(B, -1, self.num_heads, int(C // self.num_heads)).transpose(1, 2)

        x = F.scaled_dot_product_attention(q, k, v, scale=self.scale)  # pylint: disable=not-callable
        x = x.transpose(1, 2).reshape(B, N, C)
        x = self.proj(x)

        return x


class NTB(nn.Module):
    """
    Next Transformer Block
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        drop_path: float,
        stride: tuple[int, int],
        sr_ratio: int,
        mlp_ratio: float,
        head_dim: int,
        mix_block_ratio: float,
    ) -> None:
        super().__init__()
        mhsa_out_channels = make_divisible(int(out_channels * mix_block_ratio), 32)
        mhca_out_channels = out_channels - mhsa_out_channels

        self.patch_embed = PatchEmbed(in_channels, mhsa_out_channels, stride)
        self.norm1 = nn.BatchNorm2d(mhsa_out_channels)
        self.e_mhsa = E_MHSA(
            mhsa_out_channels,
            mhsa_out_channels,
            head_dim=head_dim,
            sr_ratio=sr_ratio,
        )
        self.mhsa_drop_path = StochasticDepth(drop_path * mix_block_ratio, mode="row")

        self.projection = PatchEmbed(mhsa_out_channels, mhca_out_channels, stride=(1, 1))
        self.mhca = MHCA(mhca_out_channels, head_dim=head_dim)
        self.mhca_drop_path = StochasticDepth(drop_path * (1 - mix_block_ratio), mode="row")

        self.norm2 = nn.BatchNorm2d(out_channels)
        self.mlp = MLP(out_channels, out_channels, mlp_ratio=mlp_ratio, bias=True)
        self.mlp_drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.patch_embed(x)
        (B, C, H, W) = x.size()
        out = self.norm1(x)

        out = out.reshape(B, C, H * W).permute(0, 2, 1)
        out = self.mhsa_drop_path(self.e_mhsa(out))
        out = out.permute(0, 2, 1).reshape(B, C, H, W)
        x = x + out

        out = self.projection(x)
        out = out + self.mhca_drop_path(self.mhca(out))
        x = torch.concat([x, out], dim=1)

        out = self.norm2(x)
        x = x + self.mlp_drop_path(self.mlp(out))

        return x


class NextViT(DetectorBackbone, PreTrainEncoder, MaskedTokenRetentionMixin):
    block_group_regex = r"body\.stage(\d+)\.(\d+)"

    # pylint: disable=too-many-locals
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        stem_chs = [64, 32, 64]
        strides = [1, 2, 2, 2]
        head_dim = 32
        mix_block_ratio = 0.75
        sr_ratios = [8, 4, 2, 1]
        depths: list[int] = self.config["depths"]
        drop_path_rate: float = self.config["drop_path_rate"]

        self.stage_out_channels = [
            [96] * (depths[0]),
            [192] * (depths[1] - 1) + [256],
            [384, 384, 384, 384, 512] * (depths[2] // 5),
            [768] * (depths[3] - 1) + [1024],
        ]

        self.stage_block_types = [
            [NCB] * depths[0],
            [NCB] * (depths[1] - 1) + [NTB],
            [NCB, NCB, NCB, NCB, NTB] * (depths[2] // 5),
            [NCB] * (depths[3] - 1) + [NTB],
        ]

        self.stem = nn.Sequential(
            Conv2dNormActivation(3, stem_chs[0], kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
            Conv2dNormActivation(stem_chs[0], stem_chs[1], kernel_size=(3, 3), stride=(1, 1), padding=(1, 1)),
            Conv2dNormActivation(stem_chs[1], stem_chs[2], kernel_size=(3, 3), stride=(1, 1), padding=(1, 1)),
            Conv2dNormActivation(stem_chs[2], stem_chs[2], kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
        )

        input_channel = stem_chs[-1]
        idx = 0
        stages: OrderedDict[str, nn.Module] = OrderedDict()
        return_channels: list[int] = []
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]
        for stage_id, repeats in enumerate(depths):
            layers = []
            output_channels = self.stage_out_channels[stage_id]
            block_types = self.stage_block_types[stage_id]
            for block_id in range(repeats):
                if strides[stage_id] == 2 and block_id == 0:
                    stride = (2, 2)
                else:
                    stride = (1, 1)

                output_channel = output_channels[block_id]
                block_type = block_types[block_id]  # type: ignore[index]
                if block_type is NCB:
                    layer = NCB(
                        input_channel,
                        output_channel,
                        stride=stride,
                        drop_path=dpr[idx + block_id],
                        head_dim=head_dim,
                        mlp_ratio=3,
                    )
                    layers.append(layer)

                elif block_type is NTB:
                    layer = NTB(
                        input_channel,
                        output_channel,
                        drop_path=dpr[idx + block_id],
                        stride=stride,
                        sr_ratio=sr_ratios[stage_id],
                        mlp_ratio=2,
                        head_dim=head_dim,
                        mix_block_ratio=mix_block_ratio,
                    )
                    layers.append(layer)

                else:
                    raise ValueError("Unsupported block type")

                input_channel = output_channel

            stages[f"stage{stage_id+1}"] = nn.Sequential(*layers)
            return_channels.append(output_channel)
            idx += repeats

        self.body = nn.Sequential(stages)
        self.features = nn.Sequential(
            nn.BatchNorm2d(output_channel),
            nn.AdaptiveAvgPool2d(output_size=(1, 1)),
            nn.Flatten(1),
        )
        self.return_channels = return_channels
        self.embedding_size = output_channel
        self.classifier = self.create_classifier()

        self.stem_stride = 4
        self.stem_width = stem_chs[-1]
        self.encoding_size = output_channel

        # Weights initialization
        for m in self.modules():
            if isinstance(m, (nn.BatchNorm2d, nn.GroupNorm, nn.LayerNorm, nn.BatchNorm1d)):
                nn.init.constant_(m.weight, 1.0)
                nn.init.constant_(m.bias, 0)

            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, std=0.02)
                if hasattr(m, "bias") and m.bias is not None:
                    nn.init.constant_(m.bias, 0)

            elif isinstance(m, nn.Conv2d):
                nn.init.normal_(m.weight, std=0.02)
                if hasattr(m, "bias") and m.bias is not None:
                    nn.init.constant_(m.bias, 0)

    def detection_features(self, x: torch.Tensor) -> dict[str, torch.Tensor]:
        x = self.stem(x)

        out = {}
        for name, module in self.body.named_children():
            x = module(x)
            if name in self.return_stages:
                out[name] = x

        return out

    def freeze_stages(self, up_to_stage: int) -> None:
        for param in self.stem.parameters():
            param.requires_grad = False

        for idx, module in enumerate(self.body.children()):
            if idx >= up_to_stage:
                break

            for param in module.parameters():
                param.requires_grad = False

    def masked_encoding_retention(
        self,
        x: torch.Tensor,
        mask: torch.Tensor,
        mask_token: Optional[torch.Tensor] = None,
        return_keys: Literal["all", "features", "embedding"] = "features",
    ) -> TokenRetentionResultType:
        x = self.stem(x)
        x = mask_tensor(x, mask, patch_factor=self.max_stride // self.stem_stride, mask_token=mask_token)
        x = self.body(x)

        result: TokenRetentionResultType = {}
        if return_keys in ("all", "features"):
            result["features"] = x
        if return_keys in ("all", "embedding"):
            result["embedding"] = self.features(x)

        return result

    def forward_features(self, x: torch.Tensor) -> torch.Tensor:
        x = self.stem(x)
        return self.body(x)

    def embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = self.forward_features(x)
        return self.features(x)


registry.register_model_config("nextvit_s", NextViT, config={"depths": [3, 4, 10, 3], "drop_path_rate": 0.1})
registry.register_model_config("nextvit_b", NextViT, config={"depths": [3, 4, 20, 3], "drop_path_rate": 0.2})
registry.register_model_config("nextvit_l", NextViT, config={"depths": [3, 4, 30, 3], "drop_path_rate": 0.2})



"""
data2vec 2.0, adapted from
https://github.com/facebookresearch/fairseq/blob/main/examples/data2vec/models/data2vec2.py

Paper "Efficient Self-supervised Learning with Contextualized Target Representations for Vision, Speech and Language",
https://arxiv.org/abs/2212.07525

Changes from original:
* Target CLS is taken just from the last layer
* Replaced instance norm (1st of the IN -> AVG -> LM) with layer norm
"""

# Reference license: MIT

import copy
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from birder.common.masking import get_ids_keep
from birder.layers import LayerNorm2d
from birder.net.base import MaskedTokenOmissionMixin
from birder.net.base import MaskedTokenRetentionMixin
from birder.net.base import PreTrainEncoder
from birder.net.ssl.base import SSLBaseNet


class Decoder2d(nn.Module):
    def __init__(self, in_channels: int, embed_dim: int, kernel_size: int, num_layers: int, H: int, W: int) -> None:
        super().__init__()

        self.H = H
        self.W = W

        self.blocks = nn.ModuleList()
        self.blocks.append(
            nn.Sequential(
                nn.Conv2d(
                    in_channels,
                    embed_dim,
                    kernel_size=(kernel_size, kernel_size),
                    stride=(1, 1),
                    padding=((kernel_size - 1) // 2, (kernel_size - 1) // 2),
                    groups=16,
                ),
                LayerNorm2d(embed_dim, elementwise_affine=False),
                nn.GELU(),
            )
        )
        for _ in range(num_layers - 1):
            self.blocks.append(
                nn.Sequential(
                    nn.Conv2d(
                        embed_dim,
                        embed_dim,
                        kernel_size=(kernel_size, kernel_size),
                        stride=(1, 1),
                        padding=((kernel_size - 1) // 2, (kernel_size - 1) // 2),
                        groups=16,
                    ),
                    LayerNorm2d(embed_dim, elementwise_affine=False),
                    nn.GELU(),
                )
            )

        self.proj = nn.Linear(embed_dim, in_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, _, C) = x.size()  # B, N, C

        x = x.transpose(1, 2).reshape(B, C, self.H, self.W)

        residual = x
        for i, blk in enumerate(self.blocks):
            x = blk(x)
            if i > 0:
                x = x + residual

            residual = x

        x = x.flatten(2).transpose(1, 2)
        x = self.proj(x)

        return x


class Data2Vec2(SSLBaseNet):
    def __init__(
        self,
        backbone: PreTrainEncoder,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(backbone, config=config, size=size)
        assert self.config is not None, "must set config"
        assert isinstance(self.backbone, MaskedTokenOmissionMixin)
        assert isinstance(self.backbone, MaskedTokenRetentionMixin)

        average_top_k_layers: int = self.config["average_top_k_layers"]
        decoder_dim: int = self.config["decoder_dim"]
        decoder_kernel_size: int = self.config["decoder_kernel_size"]
        decoder_layers: int = self.config["decoder_layers"]
        clone_batch: int = self.config["clone_batch"]
        cls_loss_weight: float = self.config["cls_loss_weight"]

        self.average_top_k_layers = average_top_k_layers
        self.clone_batch = clone_batch
        self.cls_loss_weight = cls_loss_weight
        self.num_patches = (self.size[0] // self.backbone.max_stride) * (self.size[1] // self.backbone.max_stride)

        self.ema_backbone = copy.deepcopy(self.backbone)
        self.decoder = Decoder2d(
            self.backbone.embedding_size,
            decoder_dim,
            decoder_kernel_size,
            decoder_layers,
            self.size[0] // self.backbone.max_stride,
            self.size[1] // self.backbone.max_stride,
        )
        self.head = nn.Linear(self.backbone.embedding_size, self.backbone.embedding_size)

        # Weights initialization
        self.ema_backbone.load_state_dict(self.backbone.state_dict())

        nn.init.trunc_normal_(self.head.weight, std=0.02)
        if self.head.bias is not None:
            nn.init.zeros_(self.head.bias)

    def forward(  # type: ignore[override]  # pylint: disable=arguments-differ
        self, src: torch.Tensor, masks: torch.Tensor
    ) -> torch.Tensor:
        # Target Representations
        y = self.ema_backbone.masked_encoding_omission(src, return_all_features=True, return_keys="all")
        y_cls = y["embedding"]
        y = y["tokens"][:, self.ema_backbone.num_special_tokens :]
        y = y[..., -self.average_top_k_layers :]  # Take the last k layers
        y = y.permute(3, 0, 1, 2)

        y = [F.layer_norm(t.float(), t.shape[-1:]) for t in y[:-1]] + [y[-1]]
        y = sum(y) / len(y)
        y = F.layer_norm(y.float(), y.shape[-1:])

        if self.clone_batch > 1:
            y = y.repeat_interleave(self.clone_batch, 0)
            y_cls = y_cls.repeat_interleave(self.clone_batch, 0)
            src = src.repeat_interleave(self.clone_batch, 0)

        ids_keep = get_ids_keep(masks)
        x = self.backbone.masked_encoding_omission(src, ids_keep=ids_keep, return_keys="all")
        x_cls = x["embedding"]
        x = x["tokens"][:, self.backbone.num_special_tokens :]

        x_cls = self.head(x_cls)

        # Using noise instead of mask tokens
        full_sequence = (
            torch.randn(x.size(0), self.num_patches, self.backbone.embedding_size, device=x.device, dtype=x.dtype)
            * 0.02
        )
        indices_expanded = ids_keep.unsqueeze(-1).expand(-1, -1, self.backbone.embedding_size)  # (B, num_kept, d_model)
        full_sequence.scatter_(1, indices_expanded, x)
        masks = masks.bool()

        predictions = self.decoder(full_sequence)

        # Compute loss only on masked positions
        patch_loss = F.mse_loss(predictions[masks], y[masks], reduction="none").sum(dim=-1).mean()

        # CLS loss
        cls_loss = F.mse_loss(x_cls, y_cls, reduction="none").sum(dim=-1).mean() * self.cls_loss_weight

        return patch_loss + cls_loss



import math
import random
from typing import Optional

import numpy as np
import torch


# Unused, keeping as a reference
def _mask_token_omission(
    x: torch.Tensor, mask_ratio: float, kept_mask_ratio: Optional[float] = None
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Apply a 1D mask to the input tensor using the MAE (Masked Autoencoder) style masking.

    Parameters
    ----------
    x
        Tensor of shape (N, L, D), where N is the batch size, L is the sequence length, and D is the feature dimension.
    mask_ratio
        The ratio of the sequence length to be masked. This value should be between 0 and 1.
    kept_mask_ratio
        The ratio of the masked tokens to be kept. If None, it defaults to the value of mask_ratio.
        This value should be between 0 and mask_ratio.

    Returns
    -------
    A tuple containing four elements:
    - The masked input tensor of shape (N, len_keep, D), where len_keep is the length of the sequence after masking.
    - The binary mask tensor of shape (N, L), where 0 indicates kept tokens and 1 indicates masked tokens.
    - The indices of kept tokens.
    - The indices to restore the original order of the sequence after masking.

    Examples
    --------
    >>> import torch
    >>> x = torch.randn(2, 10, 5)  # Example input tensor
    >>> mask_ratio = 0.5
    >>> (x_masked, mask, ids_keep, ids_restore) = _mask_token_omission(x, mask_ratio)
    >>> print(x_masked.size())  # Should print torch.Size([2, 5, 5])
    >>> print(mask.size())  # Should print torch.Size([2, 10])
    >>> print(ids_restore.size())  # Should print torch.Size([2, 10])
    """

    if kept_mask_ratio is None:
        kept_mask_ratio = mask_ratio

    # Masking: length -> length * mask_ratio
    # Perform per-sample random masking by per-sample shuffling.
    # Per-sample shuffling is done by argsort random noise.
    (N, L, D) = x.size()  # batch, length, dim
    len_keep = int(L * (1 - mask_ratio))
    len_masked = int(L * (mask_ratio - kept_mask_ratio))

    noise = torch.rand(N, L, device=x.device)  # Noise in [0, 1]

    # Sort noise for each sample
    ids_shuffle = torch.argsort(noise, dim=1)  # Ascend: small is keep, large is remove
    ids_restore = torch.argsort(ids_shuffle, dim=1)

    # Keep the first subset
    ids_keep = ids_shuffle[:, :len_keep]
    x_masked = torch.gather(x, dim=1, index=ids_keep.unsqueeze(-1).repeat(1, 1, D))

    # Generate the binary mask: 0 is keep, 1 is remove
    mask = torch.ones([N, L], device=x.device)
    mask[:, : len_keep + len_masked] = 0

    # Un-shuffle to get the binary mask
    mask = torch.gather(mask, dim=1, index=ids_restore)

    return (x_masked, mask, ids_keep, ids_restore)


def mask_tensor(
    x: torch.Tensor,
    mask: torch.Tensor,
    channels_last: bool = False,
    patch_factor: int = 1,
    mask_token: Optional[torch.Tensor] = None,
) -> torch.Tensor:
    if channels_last is False:
        x = x.permute(0, 2, 3, 1)

    (B, H, W, _) = x.size()

    shaped_mask = mask.reshape(-1, H // patch_factor, W // patch_factor)
    shaped_mask = shaped_mask.repeat_interleave(patch_factor, axis=1).repeat_interleave(patch_factor, axis=2)
    shaped_mask = shaped_mask.unsqueeze(3).type_as(x)

    if mask_token is not None:
        mask_tokens = mask_token.expand(B, H, W, -1)
        x_masked = x * (1.0 - shaped_mask) + (mask_tokens * shaped_mask)
    else:
        x_masked = x * (1.0 - shaped_mask)

    if channels_last is False:
        x_masked = x_masked.permute(0, 3, 1, 2)

    return x_masked


def uniform_mask(
    batch_size: int,
    seq_len: int,
    mask_ratio: float,
    kept_mask_ratio: Optional[float] = None,
    device: Optional[torch.device] = None,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    if kept_mask_ratio is None:
        kept_mask_ratio = mask_ratio

    # Masking: length -> length * mask_ratio
    # Perform per-sample random masking by per-sample shuffling.
    # Per-sample shuffling is done by argsort random noise.
    len_keep = int(seq_len * (1 - mask_ratio))
    len_masked = int(seq_len * (mask_ratio - kept_mask_ratio))

    noise = torch.rand(batch_size, seq_len, device=device)  # Noise in [0, 1]

    # Sort noise for each sample
    ids_shuffle = torch.argsort(noise, dim=1)  # Ascend: small is keep, large is remove
    ids_restore = torch.argsort(ids_shuffle, dim=1)

    # Keep the first subset
    ids_keep = ids_shuffle[:, :len_keep]

    # Generate the binary mask: 0 is keep, 1 is remove
    mask = torch.ones([batch_size, seq_len], device=device)
    mask[:, : len_keep + len_masked] = 0

    # Un-shuffle to get the binary mask
    mask = torch.gather(mask, dim=1, index=ids_restore)
    # assert mask.ndim == 2

    return (mask, ids_keep, ids_restore)


def get_ids_keep(mask: torch.Tensor) -> torch.Tensor:
    B = mask.size(0)
    return (1 - mask).nonzero(as_tuple=True)[1].reshape(B, -1)


def get_random_masked_indices(mask: torch.Tensor, n: int) -> torch.Tensor:
    B = mask.size(0)
    num_masked = mask.count_nonzero().item() // B
    mask_indices_abs = mask.nonzero(as_tuple=True)[1].reshape(B, -1)
    randperm = torch.argsort(torch.rand(B, num_masked, device=mask.device))[:, :n]

    return torch.gather(mask_indices_abs, index=randperm, dim=1)


def mask_from_indices(indices: torch.Tensor, seq_len: int) -> torch.Tensor:
    """
    Indices return as 1's
    """

    B = indices.size(0)
    row_indices = torch.arange(B, device=indices.device).unsqueeze(1).expand_as(indices)

    mask = torch.zeros([B, seq_len], device=indices.device)
    mask[row_indices.flatten(), indices.flatten()] = 1

    return mask


class Masking:
    def __call__(self, batch_size: int) -> torch.Tensor:
        raise NotImplementedError


class UniformMasking(Masking):
    def __init__(self, input_size: tuple[int, int], mask_ratio: float, device: Optional[torch.device] = None) -> None:
        self.seq_len = input_size[0] * input_size[1]
        self.mask_ratio = mask_ratio
        self.device = device

    def __call__(self, batch_size: int) -> torch.Tensor:
        return uniform_mask(batch_size, self.seq_len, self.mask_ratio, self.device)[0]


class BlockMasking(Masking):
    # Adapted from: https://github.com/facebookresearch/dinov2/blob/main/dinov2/data/masking.py

    def __init__(
        self,
        input_size: tuple[int, int],
        min_num_patches: int,
        max_num_patches: int,
        min_aspect: float,
        max_aspect: float,
    ) -> None:
        self.height = input_size[0]
        self.width = input_size[1]

        self.num_patches = self.height * self.width
        self.min_num_patches = min_num_patches
        self.max_num_patches = max_num_patches

        self.log_aspect_ratio = (math.log(min_aspect), math.log(max_aspect))

    def get_shape(self) -> tuple[int, int]:
        return (self.height, self.width)

    def _mask(self, mask: torch.Tensor, max_mask_patches: int) -> int:
        # 0 is keep, 1 is remove
        delta = 0
        for _ in range(10):
            target_area = random.uniform(self.min_num_patches, max_mask_patches)
            aspect_ratio = math.exp(random.uniform(*self.log_aspect_ratio))
            h = int(round(math.sqrt(target_area * aspect_ratio)))
            w = int(round(math.sqrt(target_area / aspect_ratio)))
            if w < self.width and h < self.height:
                top = random.randint(0, self.height - h)
                left = random.randint(0, self.width - w)

                num_masked = mask[top : top + h, left : left + w].sum()

                # Overlap
                if 0 < h * w - num_masked <= max_mask_patches:
                    for i in range(top, top + h):
                        for j in range(left, left + w):
                            if mask[i, j] == 0:
                                mask[i, j] = 1
                                delta += 1

                if delta > 0:
                    break

        return delta

    def __call__(self, batch_size: int) -> torch.Tensor:
        num_masking_patches = random.randint(self.min_num_patches, self.max_num_patches)
        masks = []
        for _ in range(batch_size):
            mask = torch.zeros(*self.get_shape())
            mask_count = 0
            while mask_count < num_masking_patches:
                max_mask_patches = num_masking_patches - mask_count
                max_mask_patches = min(max_mask_patches, self.max_num_patches)

                delta = self._mask(mask, max_mask_patches)
                if delta == 0:
                    break

                mask_count += delta

            masks.append(mask.flatten())

        return torch.stack(masks, dim=0)


class RollBlockMasking(Masking):
    # Adapted from: https://github.com/facebookresearch/capi/blob/main/data.py

    def __init__(
        self,
        input_size: tuple[int, int],
        num_masking_patches: int,
        min_aspect: float = 0.5,
        max_aspect: float = 2.0,
    ):
        self.height = input_size[0]
        self.width = input_size[1]
        self.num_masking_patches = num_masking_patches
        self.log_aspect_ratio = (math.log(min_aspect), math.log(max_aspect))

    def __call__(self, batch_size: int) -> torch.Tensor:
        masks = []
        for _ in range(batch_size):
            if self.num_masking_patches == 0:
                masks.append(torch.zeros(self.height * self.width))
                continue
            if self.num_masking_patches == self.height * self.width:
                masks.append(torch.ones(self.height * self.width))
                continue

            # Sample aspect ratio, not too large or too small for image
            min_lar = max(self.log_aspect_ratio[0], np.log(self.num_masking_patches / (self.width**2)))
            max_lar = min(
                self.log_aspect_ratio[1],
                np.log(self.height**2 / (self.num_masking_patches + 1e-5)),
            )
            aspect_ratio = math.exp(random.uniform(min_lar, max_lar))

            # Use ceil so mask is >= num_masking_patches
            h = int(np.ceil(math.sqrt(self.num_masking_patches * aspect_ratio)))
            w = int(np.ceil(math.sqrt(self.num_masking_patches / aspect_ratio)))
            top = random.randint(0, self.height - h)
            left = random.randint(0, self.width - w)
            b_mask = np.zeros((self.height, self.width))
            b_mask[top : top + h, left : left + w] = 1

            # truncate ids to get exactly num_masking_patches
            ids = np.where(b_mask.flatten())[0][: self.num_masking_patches]
            mask = np.zeros((self.height, self.width)).flatten()
            mask[ids] = 1
            mask_2d = mask.reshape((self.height, self.width))

            # Roll
            shift_x = random.randint(0, mask_2d.shape[0] - 1)
            shift_y = random.randint(0, mask_2d.shape[1] - 1)
            mask = np.roll(mask_2d, (shift_x, shift_y), (0, 1))
            masks.append(torch.from_numpy(mask.flatten()))

        return torch.stack(masks, dim=0)


class InverseRollBlockMasking(RollBlockMasking):
    def __init__(
        self,
        input_size: tuple[int, int],
        num_masking_patches: int,
        min_aspect: float = 0.5,
        max_aspect: float = 2.0,
    ):
        num_masking_patches = input_size[0] * input_size[1] - num_masking_patches
        super().__init__(input_size, num_masking_patches, min_aspect, max_aspect)

    def __call__(self, batch_size: int) -> torch.Tensor:
        return 1 - super().__call__(batch_size)



import argparse
import json
import logging
import time
from pathlib import Path
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import torch
from torch.utils.data import DataLoader

from birder.common import cli
from birder.common import fs_ops
from birder.common import lib
from birder.conf import settings
from birder.data.collators.detection import inference_collate_fn
from birder.data.datasets.coco import CocoInference
from birder.data.datasets.directory import make_image_dataset
from birder.data.transforms.detection import InferenceTransform
from birder.inference.detection import infer_dataloader
from birder.model_registry import registry
from birder.net.base import DetectorBackbone
from birder.results.detection import Results
from birder.results.gui import show_detections

logger = logging.getLogger(__name__)


def save_output(
    output_path: Path, sample_paths: list[str], class_to_idx: dict[str, int], detections: list[dict[str, torch.Tensor]]
) -> None:
    detection_list = [{k: v.cpu().numpy().tolist() for k, v in detection.items()} for detection in detections]
    output = dict(zip(sample_paths, detection_list))
    output["class_to_idx"] = class_to_idx
    logger.info(f"Saving output at {output_path}")
    with open(output_path, "w", encoding="utf-8") as handle:
        json.dump(output, handle, indent=2)


# pylint: disable=too-many-locals,too-many-branches
def predict(args: argparse.Namespace) -> None:
    if args.gpu is True:
        device = torch.device("cuda")
    elif args.mps is True:
        device = torch.device("mps")
    else:
        device = torch.device("cpu")

    if args.parallel is True and torch.cuda.device_count() > 1:
        logger.info(f"Using {torch.cuda.device_count()} {device} devices")
    else:
        if args.gpu_id is not None:
            torch.cuda.set_device(args.gpu_id)

        logger.info(f"Using device {device}")

    model_dtype: torch.dtype = getattr(torch, args.model_dtype)
    amp_dtype: torch.dtype = getattr(torch, args.amp_dtype)
    network_name = lib.get_detection_network_name(
        args.network, tag=args.tag, backbone=args.backbone, backbone_tag=args.backbone_tag
    )
    (net, (class_to_idx, signature, rgb_stats, *_)) = fs_ops.load_detection_model(
        device,
        args.network,
        config=args.model_config,
        tag=args.tag,
        reparameterized=args.reparameterized,
        backbone=args.backbone,
        backbone_config=args.backbone_model_config,
        backbone_tag=args.backbone_tag,
        backbone_reparameterized=args.backbone_reparameterized,
        epoch=args.epoch,
        new_size=args.size,
        quantized=args.quantized,
        inference=True,
        pts=args.pts,
        pt2=args.pt2,
        st=args.st,
        dtype=model_dtype,
    )
    logger.debug(f"RGB stats: {rgb_stats}")

    if args.dynamic_size is True or args.max_size is not None or args.no_resize:
        net.set_dynamic_size()

    if args.fast_matmul is True or args.amp is True:
        torch.set_float32_matmul_precision("high")

    if args.compile is True:
        net = torch.compile(net)
    elif args.compile_backbone is True:
        net.backbone.detection_features = torch.compile(net.backbone.detection_features)

    if args.parallel is True and torch.cuda.device_count() > 1:
        net = torch.nn.DataParallel(net)

    if args.size is None:
        args.size = lib.get_size_from_signature(signature)
        logger.debug(f"Using size={args.size}")

    score_threshold = args.min_score

    # Set label colors
    cmap = plt.get_cmap("jet")
    color_list = []
    for c in np.linspace(0, 1, len(class_to_idx) + 1):  # Include background
        rgb = cmap(c)[0:3]
        rgb = tuple(int(x * 255) for x in rgb)
        color_list.append(rgb)

    batch_size = args.batch_size
    if args.coco_json_path is not None:
        labeled = True
        root_path = Path(args.data_path[0])
        dataset = CocoInference(
            root_path,
            args.coco_json_path,
            transforms=InferenceTransform(args.size, rgb_stats, args.dynamic_size, args.max_size, args.no_resize),
        )
        if dataset.class_to_idx != class_to_idx:
            logger.warning("Dataset class to index differs from model")
    else:
        labeled = False
        root_path = Path("")
        dataset = make_image_dataset(
            args.data_path,
            {},
            transforms=InferenceTransform(args.size, rgb_stats, args.dynamic_size, args.max_size, args.no_resize),
            return_orig_sizes=True,
        )

    num_samples = len(dataset)
    inference_loader = DataLoader(
        dataset,
        batch_size=batch_size,
        shuffle=args.shuffle,
        num_workers=args.num_workers,
        prefetch_factor=args.prefetch_factor,
        collate_fn=inference_collate_fn,
    )

    show_flag = args.show is True

    def batch_callback(
        file_paths: list[str],
        _inputs: torch.Tensor,
        detections: list[dict[str, torch.Tensor]],
        _targets: list[dict[str, Any]],
        _image_sizes: list[list[int]],
    ) -> None:
        # Show flags
        if show_flag is True:
            for img_path, detection in zip(file_paths, detections):
                show_detections(
                    str(root_path.joinpath(img_path)),
                    detection,
                    class_to_idx=class_to_idx,
                    score_threshold=score_threshold,
                    color_list=color_list,
                )

    # Sort out output file names
    epoch_str = ""
    if args.epoch is not None:
        epoch_str = f"_e{args.epoch}"

    base_output_path = f"{network_name}_{len(class_to_idx)}{epoch_str}_{args.size[0]}px_{num_samples}"
    if args.model_dtype != "float32":
        base_output_path = f"{base_output_path}_{args.model_dtype}"
    if args.prefix is not None:
        base_output_path = f"{args.prefix}_{base_output_path}"
    if args.suffix is not None:
        base_output_path = f"{base_output_path}_{args.suffix}"

    output_path = settings.RESULTS_DIR.joinpath(f"{base_output_path}_output.json")

    # Inference
    tic = time.time()
    with torch.inference_mode():
        (sample_paths, detections, targets) = infer_dataloader(
            device,
            net,
            inference_loader,
            model_dtype,
            args.amp,
            amp_dtype,
            num_samples,
            batch_callback=batch_callback,
        )

    toc = time.time()
    rate = len(dataset) / (toc - tic)
    logger.info(f"{lib.format_duration(toc-tic)} to classify {len(dataset):,} samples ({rate:.2f} samples/sec)")

    # Save output
    if args.save_output is True:
        save_output(output_path, sample_paths, class_to_idx, detections)

    # Handle results
    if labeled is True:
        results = Results(sample_paths, targets, detections, class_to_idx)
        if args.save_results is True:
            results.save(f"{base_output_path}.json")

        results.log_short_report()

    else:
        if args.save_results is True:
            logger.warning("Annotations not provided, unable to save results")


def get_args_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description="Run detection prediction on directories and/or files",
        epilog=(
            "Usage example:\n"
            "python -m birder.scripts.predict_detection --network faster_rcnn --backbone resnext_101 "
            "-e 0 data/detection_data/validation\n"
            "python predict_detection.py --network retinanet --backbone resnext_101 "
            "-e 0 --show --gpu --compile data/detection_data/training\n"
            "python predict_detection.py --network faster_rcnn --backbone resnext_101 "
            "-e 0 --min-score 0.25 --gpu --show --shuffle data/detection_data/validation\n"
            "python predict_detection.py --network faster_rcnn -t coco --backbone csp_resnet_50 "
            "--backbone-tag imagenet1k -e 0 --batch-size 1 --gpu --gpu-id 1 "
            "--coco-json-path data/detection_data/validation_annotations_coco.json data/detection_data"
        ),
        formatter_class=cli.ArgumentHelpFormatter,
    )
    parser.add_argument("-n", "--network", type=str, help="the neural network to use (i.e. faster_rcnn)")
    parser.add_argument(
        "--model-config",
        action=cli.FlexibleDictAction,
        help=(
            "override the model default configuration, accepts key-value pairs or JSON "
            "('drop_path_rate=0.2' or '{\"units\": [3, 24, 36, 3], \"dropout\": 0.2}'"
        ),
    )
    parser.add_argument("--backbone", type=str, help="the neural network to used as backbone")
    parser.add_argument(
        "--backbone-model-config",
        action=cli.FlexibleDictAction,
        help=(
            "override the backbone default configuration, accepts key-value pairs or JSON "
            "('drop_path_rate=0.2' or '{\"units\": [3, 24, 36, 3], \"dropout\": 0.2}'"
        ),
    )
    parser.add_argument("--backbone-tag", type=str, help="backbone training log tag (loading only)")
    parser.add_argument(
        "--backbone-reparameterized", default=False, action="store_true", help="load reparameterized backbone"
    )
    parser.add_argument("-e", "--epoch", type=int, metavar="N", help="model checkpoint to load")
    parser.add_argument("--quantized", default=False, action="store_true", help="load quantized model")
    parser.add_argument("-t", "--tag", type=str, help="model tag (from the training phase)")
    parser.add_argument(
        "-r", "--reparameterized", default=False, action="store_true", help="load reparameterized model"
    )
    parser.add_argument("--pts", default=False, action="store_true", help="load torchscript network")
    parser.add_argument("--pt2", default=False, action="store_true", help="load standardized model")
    parser.add_argument("--st", "--safetensors", default=False, action="store_true", help="load Safetensors weights")
    parser.add_argument("--compile", default=False, action="store_true", help="enable compilation")
    parser.add_argument(
        "--compile-backbone", default=False, action="store_true", help="enable backbone only compilation"
    )
    parser.add_argument(
        "--model-dtype",
        type=str,
        choices=["float32", "float16", "bfloat16"],
        default="float32",
        help="model dtype to use",
    )
    parser.add_argument(
        "--amp", default=False, action="store_true", help="use torch.amp.autocast for mixed precision inference"
    )
    parser.add_argument(
        "--amp-dtype",
        type=str,
        choices=["float16", "bfloat16"],
        default="float16",
        help="whether to use float16 or bfloat16 for mixed precision",
    )
    parser.add_argument(
        "--fast-matmul", default=False, action="store_true", help="use fast matrix multiplication (affects precision)"
    )
    parser.add_argument("--min-score", type=float, default=0.5, help="prediction score threshold")
    parser.add_argument(
        "--size",
        type=int,
        nargs="+",
        metavar=("H", "W"),
        help=(
            "target image size as [height, width], if --dynamic-size is enabled, "
            "uses the smaller dimension as target size while preserving aspect ratio (defaults to model's signature)"
        ),
    )
    parser.add_argument(
        "--max-size",
        type=int,
        help="maximum size for the longer edge of resized images, when specified, enables dynamic sizing",
    )
    parser.add_argument(
        "--dynamic-size",
        default=False,
        action="store_true",
        help="allow variable image sizes while preserving aspect ratios",
    )
    parser.add_argument(
        "--no-resize", default=False, action="store_true", help="process images at original size without resizing"
    )
    parser.add_argument("--batch-size", type=int, default=8, metavar="N", help="the batch size")
    parser.add_argument("-j", "--num-workers", type=int, default=4, metavar="N", help="number of preprocessing workers")
    parser.add_argument(
        "--prefetch-factor", type=int, metavar="N", help="number of batches loaded in advance by each worker"
    )
    parser.add_argument("--show", default=False, action="store_true", help="show image predictions")
    parser.add_argument("--shuffle", default=False, action="store_true", help="predict samples in random order")
    parser.add_argument("--save-results", default=False, action="store_true", help="save results object")
    parser.add_argument("--save-output", default=False, action="store_true", help="save raw output as CSV")
    parser.add_argument("--prefix", type=str, help="add prefix to output file")
    parser.add_argument("--suffix", type=str, help="add suffix to output file")
    parser.add_argument("--gpu", default=False, action="store_true", help="use gpu")
    parser.add_argument("--gpu-id", type=int, metavar="ID", help="gpu id to use (ignored in parallel mode)")
    parser.add_argument("--mps", default=False, action="store_true", help="use mps (Metal Performance Shaders) device")
    parser.add_argument("--parallel", default=False, action="store_true", help="use multiple gpus")
    parser.add_argument("--coco-json-path", type=str, help="COCO json path")
    parser.add_argument("data_path", nargs="+", help="data files path (directories and files)")

    return parser


def validate_args(args: argparse.Namespace) -> None:
    args.size = cli.parse_size(args.size)

    if args.network is None:
        raise cli.ValidationError("--network is required")
    if args.backbone is None:
        raise cli.ValidationError("--backbone is required")
    if registry.exists(args.backbone, net_type=DetectorBackbone) is False:
        raise cli.ValidationError(
            f"--backbone {args.network} not supported, see list-models tool for available options"
        )
    if args.min_score >= 1 or args.min_score <= 0.0:
        raise cli.ValidationError(f"--min-score must be in range of (0, 1.0), got {args.min_score}")
    if args.parallel is True and args.gpu is False:
        raise cli.ValidationError("--parallel requires --gpu to be set")
    if args.parallel is True and args.compile is True:
        raise cli.ValidationError("--parallel cannot be used with --compile")
    if args.compile is True and args.compile_backbone is True:
        raise cli.ValidationError("--compile cannot be used with --compile-backbone")
    if args.amp is True and args.model_dtype != "float32":
        raise cli.ValidationError("--amp can only be used with --model-dtype float32")
    if args.coco_json_path is not None and len(args.data_path) > 1:
        raise cli.ValidationError(f"--coco-json-path can have at most 1 --data-path, got {len(args.data_path)}")


def args_from_dict(**kwargs: Any) -> argparse.Namespace:
    parser = get_args_parser()
    parser.set_defaults(**kwargs)
    args = parser.parse_args([])
    validate_args(args)

    return args


def main() -> None:
    parser = get_args_parser()
    args = parser.parse_args()
    validate_args(args)

    if settings.RESULTS_DIR.exists() is False:
        logger.info(f"Creating {settings.RESULTS_DIR} directory...")
        settings.RESULTS_DIR.mkdir(parents=True)

    predict(args)


if __name__ == "__main__":
    logger = logging.getLogger(getattr(__spec__, "name", __name__))
    main()



from birder.model_registry import manifest
from birder.model_registry.model_registry import Task
from birder.model_registry.model_registry import registry

__all__ = [
    "manifest",
    "Task",
    "registry",
]



"""
Self-Supervised Descriptor for Image Copy Detection (SSCD), adapted from
https://github.com/facebookresearch/sscd-copy-detection/blob/main/sscd/models/model.py

Paper "A Self-Supervised Descriptor for Image Copy Detection",
https://arxiv.org/abs/2202.10261
"""

# Reference license: MIT

from functools import partial
from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn

from birder.common import training_utils
from birder.layers import FixedGeMPool2d
from birder.model_registry import registry
from birder.net.base import BaseNet
from birder.net.ssl.base import SSLBaseNet


class SSCD(SSLBaseNet):
    def __init__(
        self,
        backbone: BaseNet,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(backbone, config=config, size=size)
        assert self.config is None, "config not supported"

        fixed_gem_pool_3: type[nn.Module] = partial(FixedGeMPool2d, 3.0)  # type: ignore[assignment]
        self.backbone = training_utils.replace_module(self.backbone, nn.AdaptiveAvgPool2d, fixed_gem_pool_3)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.backbone(x)
        x = F.normalize(x)

        return x


# Registered as classification model instead of SSL for cleaner API - functionally equivalent
registry.register_weights(
    "sscd_resnext_101_c1",
    {
        "url": "https://huggingface.co/birder-project/sscd_resnext_101_c1/resolve/main",
        "description": (
            "SSCD ResNeXt 101 model trained DISC for image copy detection. "
            "This model has not been fine-tuned for a specific classification task"
        ),
        "resolution": (320, 320),
        "formats": {
            "pts": {
                "file_size": 169.9,
                "sha256": "8ccedaf9efc243be81d8a3d432d4bd8688fb91e71b43934f68557f45bb0adb3f",
            }
        },
        "net": {"network": "resnext_101", "tag": "c1_sscd"},
    },
)



"""
CrossViT, adapted from
https://github.com/IBM/CrossViT/blob/main/models/crossvit.py

Paper "CrossViT: Cross-Attention Multi-Scale Vision Transformer for Image Classification",
https://arxiv.org/abs/2103.14899

Changes from original:
* Removed resize per patch (224/240)
* Added dynamic size support (must be a multiply of all patches, 48 by default)
"""

# Reference license: Apache-2.0

from typing import Any
from typing import Optional

import torch
import torch.nn.functional as F
from torch import nn
from torchvision.ops import StochasticDepth

from birder.model_registry import registry
from birder.net.base import BaseNet
from birder.net.vit import EncoderBlock
from birder.net.vit import adjust_position_embedding


class PatchEmbed(nn.Module):
    def __init__(self, patch_size: tuple[int, int], in_channels: int, embed_dim: int, multi_conv: bool) -> None:
        super().__init__()
        if multi_conv is True:
            if patch_size[0] == 12:
                self.proj = nn.Sequential(
                    nn.Conv2d(in_channels, embed_dim // 4, kernel_size=(7, 7), stride=(4, 4), padding=(3, 3)),
                    nn.ReLU(inplace=True),
                    nn.Conv2d(embed_dim // 4, embed_dim // 2, kernel_size=(3, 3), stride=(3, 3), padding=(0, 0)),
                    nn.ReLU(inplace=True),
                    nn.Conv2d(embed_dim // 2, embed_dim, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1)),
                )
            elif patch_size[0] == 16:
                self.proj = nn.Sequential(
                    nn.Conv2d(in_channels, embed_dim // 4, kernel_size=(7, 7), stride=(4, 4), padding=(3, 3)),
                    nn.ReLU(inplace=True),
                    nn.Conv2d(embed_dim // 4, embed_dim // 2, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
                    nn.ReLU(inplace=True),
                    nn.Conv2d(embed_dim // 2, embed_dim, kernel_size=(3, 3), stride=(2, 2), padding=(1, 1)),
                )
            else:
                raise ValueError("Unsupported patch size")

        else:
            self.proj = nn.Conv2d(in_channels, embed_dim, kernel_size=patch_size, stride=patch_size, padding=(0, 0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.proj(x)
        x = x.flatten(2).transpose(1, 2)  # NCHW -> NLC

        return x


class CrossAttention(nn.Module):
    def __init__(self, dim: int, num_heads: int, qkv_bias: bool, attn_drop: float, proj_drop: float) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.attn_drop = attn_drop
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.wq = nn.Linear(dim, dim, bias=qkv_bias)
        self.wk = nn.Linear(dim, dim, bias=qkv_bias)
        self.wv = nn.Linear(dim, dim, bias=qkv_bias)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        (B, N, C) = x.shape
        # B1C -> B1H(C/H) -> BH1(C/H)
        q = self.wq(x[:, 0:1, ...]).reshape(B, 1, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)
        # BNC -> BNH(C/H) -> BHN(C/H)
        k = self.wk(x).reshape(B, N, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)
        # BNC -> BNH(C/H) -> BHN(C/H)
        v = self.wv(x).reshape(B, N, self.num_heads, C // self.num_heads).permute(0, 2, 1, 3)

        x = F.scaled_dot_product_attention(  # pylint: disable=not-callable
            q, k, v, dropout_p=self.attn_drop if self.training else 0.0, scale=self.scale
        )
        x = x.transpose(1, 2).reshape(B, 1, C)  # (BH1N @ BHN(C/H)) -> BH1(C/H) -> B1H(C/H) -> B1C
        x = self.proj(x)
        x = self.proj_drop(x)

        return x


class CrossAttentionBlock(nn.Module):
    def __init__(
        self, dim: int, num_heads: int, qkv_bias: bool, proj_drop: float, attn_drop: float, drop_path: float
    ) -> None:
        super().__init__()
        self.norm1 = nn.LayerNorm(dim)
        self.attn = CrossAttention(
            dim, num_heads=num_heads, qkv_bias=qkv_bias, attn_drop=attn_drop, proj_drop=proj_drop
        )
        self.drop_path = StochasticDepth(drop_path, mode="row")

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x[:, 0:1, ...] + self.drop_path(self.attn(self.norm1(x)))
        return x


class MultiScaleBlock(nn.Module):
    def __init__(
        self,
        dim: list[int],
        depth: list[int],
        num_heads: list[int],
        mlp_ratio: list[float],
        qkv_bias: bool,
        proj_drop: float,
        attn_drop: float,
        drop_path: list[float],
    ) -> None:
        super().__init__()

        num_branches = len(dim)
        self.num_branches = num_branches
        self.blocks = nn.ModuleList()
        for d in range(num_branches):
            layers = []
            for i in range(depth[d]):
                layers.append(
                    EncoderBlock(
                        num_heads=num_heads[d],
                        hidden_dim=dim[d],
                        mlp_dim=int(mlp_ratio[d] * dim[d]),
                        dropout=proj_drop,
                        attention_dropout=attn_drop,
                        drop_path=drop_path[i],
                        activation_layer=nn.GELU,
                    )
                )

            self.blocks.append(nn.Sequential(*layers))

        self.projs = nn.ModuleList()
        for d in range(num_branches):
            self.projs.append(
                nn.Sequential(
                    nn.LayerNorm(dim[d]),
                    nn.GELU(),
                    nn.Linear(dim[d], dim[(d + 1) % num_branches]),
                )
            )

        self.fusion = nn.ModuleList()
        for d in range(num_branches):
            d_ = (d + 1) % num_branches
            nh = num_heads[d_]
            if depth[-1] == 0:
                self.fusion.append(
                    CrossAttentionBlock(
                        dim=dim[d_],
                        num_heads=nh,
                        qkv_bias=qkv_bias,
                        proj_drop=proj_drop,
                        attn_drop=attn_drop,
                        drop_path=drop_path[-1],
                    )
                )
            else:
                layers = []
                for _ in range(depth[-1]):
                    layers.append(
                        CrossAttentionBlock(
                            dim=dim[d_],
                            num_heads=nh,
                            qkv_bias=qkv_bias,
                            proj_drop=proj_drop,
                            attn_drop=attn_drop,
                            drop_path=drop_path[-1],
                        )
                    )

                self.fusion.append(nn.Sequential(*layers))

        self.revert_projs = nn.ModuleList()
        for d in range(num_branches):
            self.revert_projs.append(
                nn.Sequential(
                    nn.LayerNorm(dim[(d + 1) % num_branches]),
                    nn.GELU(),
                    nn.Linear(dim[(d + 1) % num_branches], dim[d]),
                )
            )

    def forward(self, x: list[torch.Tensor]) -> list[torch.Tensor]:
        outs_b = []
        for i, block in enumerate(self.blocks):
            outs_b.append(block(x[i]))

        proj_cls_token: list[torch.Tensor] = []
        for i, proj in enumerate(self.projs):
            proj_cls_token.append(proj(outs_b[i][:, 0:1, ...]))

        # Cross attention
        outs = []
        for i, (fusion, revert_proj) in enumerate(zip(self.fusion, self.revert_projs)):
            tmp = torch.concat((proj_cls_token[i], outs_b[(i + 1) % self.num_branches][:, 1:, ...]), dim=1)
            tmp = fusion(tmp)
            reverted_proj_cls_token = revert_proj(tmp[:, 0:1, ...])
            tmp = torch.concat((reverted_proj_cls_token, outs_b[i][:, 1:, ...]), dim=1)
            outs.append(tmp)

        return outs


def _compute_num_patches(img_size: list[tuple[int, int]], patches: list[int]) -> list[int]:
    return [(i[0] // p) * (i[1] // p) for i, p in zip(img_size, patches)]


class CrossViT(BaseNet):
    default_size = (240, 240)

    # pylint: disable=too-many-locals
    def __init__(
        self,
        input_channels: int,
        num_classes: int,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__(input_channels, num_classes, config=config, size=size)
        assert self.config is not None, "must set config"

        qkv_bias = True
        pos_drop_rate = 0.0
        proj_drop_rate = 0.0
        attn_drop_rate = 0.0
        patch_size = [12, 16]
        embed_dim: list[int] = self.config["embed_dim"]
        depths: list[list[int]] = self.config["depths"]
        num_heads: list[int] = self.config["num_heads"]
        mlp_ratio: list[float] = self.config["mlp_ratio"]
        multi_conv: bool = self.config["multi_conv"]
        drop_path_rate: float = self.config["drop_path_rate"]

        image_size = [(self.size[0], self.size[1])] * len(patch_size)
        num_patches = _compute_num_patches(image_size, patch_size)
        self.patch_size = patch_size
        self.num_branches = len(patch_size)
        self.embed_dim = embed_dim

        self.pos_embed = nn.ParameterList(
            [nn.Parameter(torch.zeros(1, 1 + num_patches[i], embed_dim[i])) for i in range(self.num_branches)]
        )
        self.cls_token = nn.ParameterList(
            [nn.Parameter(torch.zeros(1, 1, embed_dim[i])) for i in range(self.num_branches)]
        )

        self.patch_embed = nn.ModuleList()
        for p, d in zip(patch_size, embed_dim):
            self.patch_embed.append(
                PatchEmbed(
                    patch_size=(p, p),
                    in_channels=self.input_channels,
                    embed_dim=d,
                    multi_conv=multi_conv,
                )
            )

        self.pos_drop = nn.Dropout(p=pos_drop_rate)

        total_depth = sum([sum(x[-2:]) for x in depths])  # pylint: disable=consider-using-generator
        dpr = [x.item() for x in torch.linspace(0, drop_path_rate, total_depth)]  # Stochastic depth decay rule
        dpr_ptr = 0
        self.blocks = nn.ModuleList()
        for depth in depths:
            curr_depth = max(depth[:-1]) + depth[-1]
            block = MultiScaleBlock(
                embed_dim,
                depth,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                proj_drop=proj_drop_rate,
                attn_drop=attn_drop_rate,
                drop_path=dpr[dpr_ptr : dpr_ptr + curr_depth],
            )
            dpr_ptr += curr_depth
            self.blocks.append(block)

        self.norm = nn.ModuleList([nn.LayerNorm(embed_dim[i]) for i in range(self.num_branches)])
        self.embedding_size = sum(self.embed_dim)
        self.classifier = nn.ModuleList()
        for i in range(self.num_branches):
            self.classifier.append(self.create_classifier(self.embed_dim[i]))

        # Weights initialization
        for m in self.modules():
            if isinstance(m, nn.Linear):
                nn.init.trunc_normal_(m.weight, std=0.02)
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

            elif isinstance(m, nn.LayerNorm):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)

    def reset_classifier(self, num_classes: int) -> None:
        self.num_classes = num_classes
        self.classifier = nn.ModuleList()
        for i in range(self.num_branches):
            self.classifier.append(self.create_classifier(self.embed_dim[i]))

    def forward_features(self, x: torch.Tensor) -> list[torch.Tensor]:
        B = x.shape[0]
        xs = []
        for patch_embed, cls_tokens, pos_embed in zip(self.patch_embed, self.cls_token, self.pos_embed):
            branch = patch_embed(x)
            cls_tokens = cls_tokens.expand(B, -1, -1)
            branch = torch.concat((cls_tokens, branch), dim=1)
            branch = branch + pos_embed
            branch = self.pos_drop(branch)
            xs.append(branch)

        for block in self.blocks:
            xs = block(xs)

        xs = [norm(xs[i]) for i, norm in enumerate(self.norm)]

        return xs

    def embedding(self, x: torch.Tensor) -> list[torch.Tensor]:
        xs = self.forward_features(x)
        out = [x[:, 0] for x in xs]

        return out

    def classify(self, x: list[torch.Tensor]) -> torch.Tensor:
        x = [classifier(x[i]) for i, classifier in enumerate(self.classifier)]
        out = torch.mean(torch.stack(x, dim=0), dim=0)

        return out

    def set_dynamic_size(self, dynamic_size: bool = True) -> None:
        assert dynamic_size is False, "Dynamic size not supported for this network"

    def adjust_size(self, new_size: tuple[int, int]) -> None:
        if new_size == self.size:
            return

        old_size = self.size
        super().adjust_size(new_size)

        # Sort out sizes
        for i in range(self.num_branches):
            old_h = old_size[0] // self.patch_size[i]
            old_w = old_size[1] // self.patch_size[i]
            h = new_size[0] // self.patch_size[i]
            w = new_size[1] // self.patch_size[i]
            self.pos_embed[i] = nn.Parameter(
                adjust_position_embedding(self.pos_embed[i], (old_h, old_w), (h, w), num_prefix_tokens=1)
            )


registry.register_model_config(
    "crossvit_t",
    CrossViT,
    config={
        "embed_dim": [96, 192],
        "depths": [[1, 4, 0], [1, 4, 0], [1, 4, 0]],
        "num_heads": [3, 3],
        "mlp_ratio": [4.0, 4.0, 1.0],
        "multi_conv": False,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "crossvit_9",
    CrossViT,
    config={
        "embed_dim": [128, 256],
        "depths": [[1, 3, 0], [1, 3, 0], [1, 3, 0]],
        "num_heads": [4, 4],
        "mlp_ratio": [3.0, 3.0, 1.0],
        "multi_conv": False,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "crossvit_9d",
    CrossViT,
    config={
        "embed_dim": [128, 256],
        "depths": [[1, 3, 0], [1, 3, 0], [1, 3, 0]],
        "num_heads": [4, 4],
        "mlp_ratio": [3.0, 3.0, 1.0],
        "multi_conv": True,
        "drop_path_rate": 0.0,
    },
)
registry.register_model_config(
    "crossvit_s",
    CrossViT,
    config={
        "embed_dim": [192, 384],
        "depths": [[1, 4, 0], [1, 4, 0], [1, 4, 0]],
        "num_heads": [6, 6],
        "mlp_ratio": [4.0, 4.0, 1.0],
        "multi_conv": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "crossvit_15",
    CrossViT,
    config={
        "embed_dim": [192, 384],
        "depths": [[1, 5, 0], [1, 5, 0], [1, 5, 0]],
        "num_heads": [6, 6],
        "mlp_ratio": [3.0, 3.0, 1.0],
        "multi_conv": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "crossvit_15d",
    CrossViT,
    config={
        "embed_dim": [192, 384],
        "depths": [[1, 5, 0], [1, 5, 0], [1, 5, 0]],
        "num_heads": [6, 6],
        "mlp_ratio": [3.0, 3.0, 1.0],
        "multi_conv": True,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "crossvit_b",
    CrossViT,
    config={
        "embed_dim": [384, 768],
        "depths": [[1, 4, 0], [1, 4, 0], [1, 4, 0]],
        "num_heads": [12, 12],
        "mlp_ratio": [4.0, 4.0, 1.0],
        "multi_conv": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "crossvit_18",
    CrossViT,
    config={
        "embed_dim": [224, 448],
        "depths": [[1, 6, 0], [1, 6, 0], [1, 6, 0]],
        "num_heads": [7, 7],
        "mlp_ratio": [3.0, 3.0, 1.0],
        "multi_conv": False,
        "drop_path_rate": 0.1,
    },
)
registry.register_model_config(
    "crossvit_18d",
    CrossViT,
    config={
        "embed_dim": [224, 448],
        "depths": [[1, 6, 0], [1, 6, 0], [1, 6, 0]],
        "num_heads": [7, 7],
        "mlp_ratio": [3.0, 3.0, 1.0],
        "multi_conv": True,
        "drop_path_rate": 0.1,
    },
)

registry.register_weights(
    "crossvit_9d_il-common",
    {
        "description": "CrossViT 9 dagger model trained on the il-common dataset",
        "resolution": (240, 240),
        "formats": {
            "pt": {
                "file_size": 32.7,
                "sha256": "515265ed725adce09464bfd23ce612b1d1178bc22a57960db089d7148556149a",
            }
        },
        "net": {"network": "crossvit_9d", "tag": "il-common"},
    },
)



from typing import Any
from typing import Optional
from typing import TypedDict

import torch
from torch import nn

from birder.model_registry import Task
from birder.model_registry import registry
from birder.net.base import BaseNet
from birder.net.base import DataShapeType

SSLSignatureType = TypedDict(
    "SSLSignatureType",
    {
        "inputs": list[DataShapeType],
        "outputs": list[DataShapeType],
    },
)


def get_ssl_signature(input_shape: tuple[int, ...]) -> SSLSignatureType:
    return {
        "inputs": [{"data_shape": [0, *input_shape[1:]]}],
        "outputs": [{"data_shape": [0]}],
    }


class SSLBaseNet(nn.Module):
    auto_register = True
    square_only = False
    task = str(Task.SELF_SUPERVISED_LEARNING)

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if cls.auto_register is False:
            # Exclude networks with custom config (initialized only with aliases)
            return

        registry.register_model(cls.__name__.lower(), cls)

    def __init__(
        self,
        backbone: BaseNet,
        *,
        config: Optional[dict[str, Any]] = None,
        size: Optional[tuple[int, int]] = None,
    ) -> None:
        super().__init__()
        self.input_channels = backbone.input_channels
        self.backbone = backbone

        if hasattr(self, "config") is False:  # Avoid overriding aliases
            self.config = config
        elif config is not None:
            assert self.config is not None
            self.config.update(config)  # Override with custom config

        if size is not None:
            self.size = size
        else:
            self.size = self.backbone.size

        assert isinstance(self.size, tuple)
        assert isinstance(self.size[0], int)
        assert isinstance(self.size[1], int)
        if self.square_only is True:
            assert self.size[0] == self.size[1]

        self.backbone.adjust_size(self.size)

    def forward(self, x: torch.Tensor) -> Any:
        raise NotImplementedError



from birder.layers.activations import QuickGELU
from birder.layers.attention_pool import MultiHeadAttentionPool
from birder.layers.ffn import FFN
from birder.layers.ffn import SwiGLU_FFN
from birder.layers.gem import FixedGeMPool2d
from birder.layers.gem import GeMPool2d
from birder.layers.layer_norm import LayerNorm2d
from birder.layers.layer_scale import LayerScale
from birder.layers.layer_scale import LayerScale2d

__all__ = [
    "QuickGELU",
    "MultiHeadAttentionPool",
    "FFN",
    "SwiGLU_FFN",
    "FixedGeMPool2d",
    "GeMPool2d",
    "LayerNorm2d",
    "LayerScale",
    "LayerScale2d",
]
